"""
snidtype.py – post-processing / typing / statistics for the Python SNID port
============================================================================

This module implements the exact type determination logic from Fortran SNID,
including template ranking, type fractions, slopes, and security assessment.

Public API
----------
SNIDResult                       *dataclass container* returned by run_snid
determine_best_type_fortran      *main function* - complete Fortran pipeline
compute_type_fractions           weighted fractions of each SN type
compute_subtype_fractions        same for sub-types
compute_consensus_z_age          ensemble ⟨z⟩, ⟨age⟩ with covariance
"""

from __future__ import annotations
import numpy as np
from dataclasses import dataclass, field
from typing import Dict, List, Tuple, Any, Set, Optional
from collections import Counter, defaultdict
import logging
from snid_sage.shared.utils.math_utils.weighted_statistics import (
    estimate_weighted_redshift,
    weighted_redshift_error,
    estimate_weighted_epoch,
    weighted_epoch_error,
)

# ----------------------------------------------------------------------
# 0.  Constants from Fortran SNID.INC
# ----------------------------------------------------------------------
RLAP_MIN   = 5.0
LAP_MIN    = 0.4
# Z_FILTER constant removed (deprecated)
EPSFRAC    = 0.01
EPSSLOPE   = 0.01
EPSRLAP    = 0.01  # For multiple best template check
MAXRLAP    = 100   # Maximum rlap value for slope calculations

# DEPRECATED: Legacy RLAP security thresholds removed
# Use cluster-based security assessment from cosmological_clustering.py instead

# Fortran type mapping (from enhanced typeinfo.f - 13 types, 70+ subtypes)
TYPENAME = {
    # Type 1 - SN Ia (10 subtypes)
    1: {1: 'Ia', 2: 'Ia-norm', 3: 'Ia-91T', 4: 'Ia-91bg', 5: 'Ia-csm', 6: 'Ia-pec', 
        7: 'Ia-02cx', 8: 'Ia-03fg', 9: 'Ia-02es', 10: 'Ia-Ca-rich'},
    
    # Type 2 - SN Ib (7 subtypes)
    2: {1: 'Ib', 2: 'Ib-norm', 3: 'Ib-pec', 4: 'IIb', 5: 'Ibn', 6: 'Ib-Ca-rich', 7: 'Ib-csm'},
    
    # Type 3 - SN Ic (7 subtypes)
    3: {1: 'Ic', 2: 'Ic-norm', 3: 'Ic-pec', 4: 'Ic-broad', 5: 'Icn', 6: 'Ic-Ca-rich', 7: 'Ic-csm'},
    
    # Type 4 - SN II (6 subtypes)
    4: {1: 'II', 2: 'IIP', 3: 'II-pec', 4: 'IIn', 5: 'IIL', 6: 'IIn-pec'},
    
    # Type 5 - NotSN (9 subtypes) - LEGACY, kept for backward compatibility
    5: {1: 'NotSN', 2: 'AGN', 3: 'Gal', 4: 'QSO', 5: 'M-star', 6: 'C-star', 
        7: 'Afterglow', 8: 'Nova', 9: 'CV'},
    
    # Type 6 - SLSN (6 subtypes)
    6: {1: 'SLSN', 2: 'SLSN-I', 3: 'SLSN-Ib', 4: 'SLSN-Ic', 5: 'SLSN-II', 6: 'SLSN-IIn'},
    
    # Type 7 - LFBOT (3 subtypes)
    7: {1: 'LFBOT', 2: '18cow', 3: '20xnd'},
    
    # Type 8 - TDE (5 subtypes)
    8: {1: 'TDE', 2: 'TDE-H', 3: 'TDE-He', 4: 'TDE-H-He', 5: 'TDE-Ftless'},
    
    # Type 9 - KN (2 subtypes)
    9: {1: 'KN', 2: '17gfo'},
    
    # Type 10 - GAP (4 subtypes)
    10: {1: 'GAP', 2: 'LRN', 3: 'LBV', 4: 'ILRT'},
    
    # Type 11 - Galaxy (7 subtypes)
    11: {1: 'Galaxy', 2: 'Gal-E', 3: 'Gal-S0', 4: 'Gal-Sa', 5: 'Gal-Sb', 
         6: 'Gal-Sc', 7: 'Gal-SB'},
    
    # Type 12 - Star (3 subtypes)
    12: {1: 'Star', 2: 'M-star', 3: 'C-star'},
    
    # Type 13 - AGN (3 subtypes)
    13: {1: 'AGN', 2: 'AGN-type1', 3: 'QSO'}
}

# Reverse mapping (string to type indices)
TYPE_TO_INDICES = {}
for it, subtypes in TYPENAME.items():
    for ist, name in subtypes.items():
        TYPE_TO_INDICES[name] = (it, ist)

# Flat type to hierarchical main type mapping (for templates using flat structure)
FLAT_TO_MAIN_TYPE = {
    # Core types that are already main types
    'Ia': 'Ia', 'Ib': 'Ib', 'Ic': 'Ic', 'II': 'II', 'NotSN': 'NotSN',
    
    # SLSN subtypes → SLSN main type
    'SLSN': 'SLSN', 'SLSN-I': 'SLSN', 'SLSN-Ib': 'SLSN', 'SLSN-Ic': 'SLSN', 
    'SLSN-II': 'SLSN', 'SLSN-IIn': 'SLSN',
    
    # LFBOT subtypes → LFBOT main type
    'LFBOT': 'LFBOT', '18cow': 'LFBOT', '20xnd': 'LFBOT',
    
    # TDE subtypes → TDE main type
    'TDE': 'TDE', 'TDE-H': 'TDE', 'TDE-He': 'TDE', 'TDE-H-He': 'TDE', 'TDE-Ftless': 'TDE',
    
    # KN subtypes → KN main type
    'KN': 'KN', '17gfo': 'KN',
    
    # GAP subtypes → GAP main type (but often used as main types in templates)
    'GAP': 'GAP', 'LRN': 'GAP', 'LBV': 'GAP', 'ILRT': 'GAP',
    
    # NEW: Galaxy subtypes → Galaxy main type
    'Galaxy': 'Galaxy', 'Gal': 'Galaxy', 'Gal-E': 'Galaxy', 'Gal-S0': 'Galaxy',
    'Gal-Sa': 'Galaxy', 'Gal-Sb': 'Galaxy', 'Gal-Sc': 'Galaxy', 'Gal-SB': 'Galaxy',
    
    # NEW: Star subtypes → Star main type
    'Star': 'Star', 'M-star': 'Star', 'C-star': 'Star',
    
    # NEW: AGN subtypes → AGN main type
    'AGN': 'AGN', 'AGN-type1': 'AGN', 'QSO': 'AGN',
    
    # NotSN subtypes (LEGACY - for backward compatibility)
    'Afterglow': 'GAP', 'Nova': 'GAP', 'CV': 'Star',
    
    # Special cases found in your templates
    'US-Ic': 'Ic'  # Ultra-stripped Type Ic
}

def get_main_type_from_template(template_type: str) -> str:
    """
    Convert template type (which may be flat) to hierarchical main type.
    
    Parameters
    ----------
    template_type : str
        Type from template (e.g., 'SLSN-I', 'TDE-He', 'Ia-norm')
        
    Returns
    -------
    str
        Main type for classification (e.g., 'SLSN', 'TDE', 'Ia')
    """
    # Handle hierarchical subtypes (e.g., 'Ia-norm' → 'Ia')
    if template_type.startswith(('Ia-', 'Ib-', 'Ic-', 'II')):
        if template_type.startswith('Ia-'):
            return 'Ia'
        elif template_type.startswith('Ib-'):
            return 'Ib'
        elif template_type.startswith('Ic-'):
            return 'Ic'
        elif template_type.startswith('II'):
            return 'II'
    
    # Use flat-to-main mapping
    return FLAT_TO_MAIN_TYPE.get(template_type, template_type)

def get_type_indices_from_template(template_type: str, template_subtype: str = "") -> Tuple[int, int]:
    """
    Get type indices for a template, handling both hierarchical and flat structures.
    
    Parameters
    ----------
    template_type : str
        Main type from template
    template_subtype : str, optional
        Subtype from template (if available)
        
    Returns
    -------
    Tuple[int, int]
        (main_type_index, subtype_index) for TYPENAME lookup
    """
    # Use subtype if available, otherwise use type
    lookup_name = template_subtype if template_subtype else template_type
    
    # Try direct lookup first
    if lookup_name in TYPE_TO_INDICES:
        return TYPE_TO_INDICES[lookup_name]
    
    # Try to find by main type classification
    main_type = get_main_type_from_template(template_type)
    
    # Look for main type in TYPENAME
    for it, subtypes in TYPENAME.items():
        if subtypes.get(1) == main_type:  # Check if this is the main type
            # Default to main type (ist=1) if subtype not found
            return (it, 1)
    
    # Fallback to unknown
    return (0, 0)

# ----------------------------------------------------------------------
# 1.  Result container
# ----------------------------------------------------------------------
@dataclass
class SNIDResult:
    """
    Container for SNID results with all match information and statistics.
    """
    # Status / bookkeeping
    success: bool = False
    runtime_sec: float = 0.0
    
    # Input spectrum data
    input_spectrum: Dict[str, np.ndarray] = field(default_factory=dict)
    processed_spectrum: Dict[str, np.ndarray] = field(default_factory=dict)
    
    # Template matches
    best_matches: List[Dict[str, Any]] = field(default_factory=list)
    filtered_matches: List[Dict[str, Any]] = field(default_factory=list)  # GMM-filtered matches
    top_matches: List[Dict[str, Any]] = field(default_factory=list)  # Top N matches for output files
    
    # Basic match parameters
    r: float = 0.0
    lap: float = 0.0
    rlap: float = 0.0
    redshift: float = 0.0
    redshift_error: float = 0.0
    template_name: str = "Unknown"
    template_type: str = "Unknown"
    template_subtype: str = ""
    template_age: float = 0.0
    
    # Consensus values
    initial_redshift: float = 0.0
    consensus_redshift: float = 0.0
    consensus_redshift_error: float = 0.0
    consensus_z_median: float = 0.0
    consensus_age: float = 0.0
    consensus_age_error: float = 0.0
    consensus_type: str = "Unknown"
    best_subtype: str = "Unknown"
    type_confidence: float = 0.0
    subtype_confidence: float = 0.0  # NEW: Probabilistic subtype confidence from cluster-aware method

    
    # Statistics
    match_statistics: Dict[str, Any] = field(default_factory=dict)
    type_fractions: Dict[str, float] = field(default_factory=dict)
    type_fractions_weighted: Dict[str, float] = field(default_factory=dict)
    type_statistics: Dict[str, Any] = field(default_factory=dict)
    
    # Additional comprehensive statistics (matching Fortran SNID)
    type_slopes: Dict[str, float] = field(default_factory=dict)
    subtype_fractions: Dict[str, Dict[str, float]] = field(default_factory=dict)
    template_rankings: Dict[str, List[int]] = field(default_factory=dict)
    correlation_statistics: Dict[str, Any] = field(default_factory=dict)
    
    # Grid parameters
    dwlog: float | None = None
    min_rlap: float | None = None
    log_wave: np.ndarray = field(default_factory=lambda: np.array([]))
    
    # Output files
    output_files: Dict[str, str] = field(default_factory=dict)
    input_file: str = ""

    def __str__(self) -> str:
        if not self.success:
            return "SNID run failed."
        
        # ENHANCED: More informative and prominent result display
        lines = []
        
        # Primary classification line with enhanced formatting
        redshift_str = f"z = {self.redshift:.5f}"
        if hasattr(self, 'redshift_error') and self.redshift_error > 0:
            redshift_str += f" ± {self.redshift_error:.5f}"
        
        quality_str = f"RLAP = {self.rlap:.2f}"
        
        # Enhanced type/subtype display with confidence
        if self.consensus_type and self.consensus_type != "Unknown":
            type_str = self.consensus_type
            
            # Add subtype with confidence if available
            if (self.best_subtype and self.best_subtype != "Unknown" and 
                hasattr(self, 'subtype_confidence') and self.subtype_confidence > 0):
                
                # Convert confidence to descriptive level (qualitative only like CLI)
                if self.subtype_confidence > 0.7:
                    conf_icon = "🔒"
                    conf_desc = "High"
                elif self.subtype_confidence > 0.4:
                    conf_icon = "🔓"
                    conf_desc = "Medium"
                else:
                    conf_icon = "🔑"
                    conf_desc = "Low"
                
                type_str += f" / {self.best_subtype} ({conf_icon} {conf_desc})"
            elif self.best_subtype and self.best_subtype != "Unknown":
                type_str += f" / {self.best_subtype}"
            

        else:
            type_str = "Unknown (❌ FAILED)"
        
        # Clustering information if available
        cluster_info = ""
        if (hasattr(self, 'clustering_results') and self.clustering_results and 
            self.clustering_results.get('success')):
            cluster = self.clustering_results.get('best_cluster')
            if cluster:
                cluster_size = cluster.get('size', 0)
                cluster_score = cluster.get('composite_score', 0)
                top_5_mean = cluster.get('top_5_mean', 0)
                cluster_info = f"   🎯 Cluster: {cluster_size} templates (top-5 mean: {top_5_mean:.2f}, score: {cluster_score:.2f})"
        
        # Consensus redshift if different from template redshift
        consensus_info = ""
        if (hasattr(self, 'consensus_redshift') and self.consensus_redshift > 0 and 
            abs(self.consensus_redshift - self.redshift) > 0.001):
            consensus_info = f"   📊 Consensus z = {self.consensus_redshift:.5f}"
            if hasattr(self, 'consensus_redshift_error') and self.consensus_redshift_error > 0:
                consensus_info += f" ± {self.consensus_redshift_error:.5f}"
        
        # Build the final string with clearer hierarchy
        lines.append("=" * 60)
        lines.append(f"🏷️  Classification: {type_str}")
        lines.append(f"🔴 {redshift_str}   ⭐ {quality_str}")
        lines.append("=" * 60)
        
        if cluster_info:
            lines.append(cluster_info)
        if consensus_info:
            lines.append(consensus_info)
        
        return "\n".join(lines)


# ----------------------------------------------------------------------
# 2.  Helper functions
# ----------------------------------------------------------------------
 


def linear_fit_weighted(x: np.ndarray, y: np.ndarray, w: np.ndarray) -> Tuple[float, float]:
    """Weighted linear regression returning slope and intercept."""
    if len(x) < 2:
        return 0.0, 0.0
        
    # Weighted fit
    sw = w.sum()
    if sw == 0:
        return 0.0, 0.0
        
    wx = w * x
    wy = w * y
    swx = wx.sum()
    swy = wy.sum()
    swxx = (wx * x).sum()
    swxy = (wx * y).sum()
    
    delta = sw * swxx - swx * swx
    if delta == 0:
        return 0.0, 0.0
        
    a0 = (swxx * swy - swx * swxy) / delta  # intercept
    a1 = (sw * swxy - swx * swy) / delta    # slope
    
    return a1, a0


# ----------------------------------------------------------------------
# 3.  Core statistics functions (following Fortran logic exactly)
# ----------------------------------------------------------------------
def compute_type_subtype_stats(matches: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    Compute redshift/age statistics by type and subtype.
    Following Fortran getzt logic.
    """
    if not matches:
        return {}
    
    # Group by type and subtype
    type_groups = {}
    for m in matches:
        tp = m['template'].get('type', 'Unknown')
        sub = m['template'].get('subtype', 'Unknown')
        
        if tp not in type_groups:
            type_groups[tp] = {'_all': [], 'subtypes': {}}
        
        type_groups[tp]['_all'].append(m)
        
        if sub not in type_groups[tp]['subtypes']:
            type_groups[tp]['subtypes'][sub] = []
        type_groups[tp]['subtypes'][sub].append(m)
    
    # Compute statistics for each group
    stats = {}
    for tp, data in type_groups.items():
        stats[tp] = {}
        
        # Type-level statistics
        all_matches = data['_all']
        z_vals = np.array([m['redshift'] for m in all_matches])
        z_errs = np.array([m.get('redshift_error', 0.0) for m in all_matches])
        # Use RLAP-cos if available, otherwise RLAP
        from snid_sage.shared.utils.math_utils import get_best_metric_value
        rlap_metrics = np.array([get_best_metric_value(m) for m in all_matches])
        
        # Weighted mean and error using canonical weights (unbiased weighted SD)
        z_mean = estimate_weighted_redshift(z_vals, z_errs, rlap_metrics)
        z_err = weighted_redshift_error(z_vals, z_errs, rlap_metrics)
        
        # Age statistics
        age_vals = []
        age_rlap_metrics = []
        age_z_errs = []
        for m in all_matches:
            age = m.get('age', m['template'].get('age', 0.0))
            age_flag = m['template'].get('age_flag', 0)
            if age_flag == 0:
                z_err = m.get('redshift_error', 0.0)
                if np.isfinite(age) and np.isfinite(z_err) and z_err > 0:
                    age_vals.append(age)
                    # Use RLAP-cos if available, otherwise RLAP
                    age_rlap_metrics.append(get_best_metric_value(m))
                    age_z_errs.append(z_err)
        
        if age_vals:
            age_vals = np.array(age_vals)
            age_rlap_metrics = np.array(age_rlap_metrics)
            age_z_errs = np.array(age_z_errs)
            age_mean = estimate_weighted_epoch(age_vals, age_z_errs, age_rlap_metrics)
            age_err = weighted_epoch_error(age_vals, age_z_errs, age_rlap_metrics)
        else:
            age_mean = age_err = 0.0
        
        stats[tp]['_all'] = {
            'z_mean': z_mean, 'z_err': z_err,
            'age_mean': age_mean, 'age_err': age_err,
            'count': len(all_matches)
        }
        
        # Subtype-level statistics
        for sub, sub_matches in data['subtypes'].items():
            if len(sub_matches) == 0:
                continue
                
            z_vals = np.array([m['redshift'] for m in sub_matches])
            z_errs = np.array([m.get('redshift_error', 0.0) for m in sub_matches])
            # Use RLAP-cos if available, otherwise RLAP
            rlap_metrics = np.array([get_best_metric_value(m) for m in sub_matches])
            
            z_mean = estimate_weighted_redshift(z_vals, z_errs, rlap_metrics)
            z_err = weighted_redshift_error(z_vals, z_errs, rlap_metrics)
            
            # Age statistics
            age_vals = []
            age_rlap_metrics = []
            age_z_errs = []
            for m in sub_matches:
                age = m.get('age', m['template'].get('age', 0.0))
                age_flag = m['template'].get('age_flag', 0)
                if age_flag == 0:
                    z_err = m.get('redshift_error', 0.0)
                    if np.isfinite(age) and np.isfinite(z_err) and z_err > 0:
                        age_vals.append(age)
                        # Use RLAP-cos if available, otherwise RLAP
                        age_rlap_metrics.append(get_best_metric_value(m))
                        age_z_errs.append(z_err)
            
            if age_vals:
                age_vals = np.array(age_vals)
                age_rlap_metrics = np.array(age_rlap_metrics)
                age_z_errs = np.array(age_z_errs)
                age_mean = estimate_weighted_epoch(age_vals, age_z_errs, age_rlap_metrics)
                age_err = weighted_epoch_error(age_vals, age_z_errs, age_rlap_metrics)
            else:
                age_mean = age_err = 0.0
            
            stats[tp][sub] = {
                'z_mean': z_mean, 'z_err': z_err,
                'age_mean': age_mean, 'age_err': age_err,
                'count': len(sub_matches)
            }
    
    return stats

# ----------------------------------------------------------------------
# 5.  Simplified API functions (for backward compatibility)
# ----------------------------------------------------------------------
def compute_type_fractions(matches: List[Dict[str, Any]], 
                          weighted: bool = False) -> Dict[str, float]:
    """Compute type fractions from matches."""
    if not matches:
        return {}
        
    type_values = defaultdict(float)
    
    for m in matches:
        tp = m['template'].get('type', 'Unknown')
        value = m['rlap'] if weighted else 1.0
        type_values[tp] += value
    
    total = sum(type_values.values())
    if total > 0:
        return {tp: val/total for tp, val in type_values.items()}
    else:
        return {}


def compute_subtype_fractions(matches: List[Dict[str, Any]], 
                             weighted: bool = True) -> Dict[str, Dict[str, float]]:
    """Compute subtype fractions from matches."""
    if not matches:
        return {}
        
    type_subtype_values = defaultdict(lambda: defaultdict(float))
    
    for m in matches:
        tp = m['template'].get('type', 'Unknown')
        sub = m['template'].get('subtype', 'Unknown')
        value = m['rlap'] if weighted else 1.0
        type_subtype_values[tp][sub] += value
    
    result = {}
    for tp, subtypes in type_subtype_values.items():
        total = sum(subtypes.values())
        if total > 0:
            result[tp] = {sub: val/total for sub, val in subtypes.items()}
        else:
            result[tp] = {sub: 0.0 for sub in subtypes}
    
    return result


 


# ----------------------------------------------------------------------
# 6.  Legacy functions (for backward compatibility)
# ----------------------------------------------------------------------
def determine_best_type(matches: List[Dict[str, Any]], 
                       min_rlap: float = RLAP_MIN) -> str:
    """Legacy function - use determine_best_type_fortran instead."""
    if not matches:
        return 'Unknown'
    
    # Simple type determination based on highest rlap
    good_matches = [m for m in matches if m['rlap'] >= min_rlap]
    if not good_matches:
        return 'Unknown'
    
    # Return type of best match
    return good_matches[0]['template'].get('type', 'Unknown')


def determine_type_security_detailed(matches: List[Dict[str, Any]], 
                                   consensus_type: str,
                                   min_rlap: float = RLAP_MIN) -> Dict[str, Any]:
    """Legacy function - use determine_best_type_fortran instead."""
    good_matches = [m for m in matches if m['rlap'] >= min_rlap]
    if not good_matches:
        return {'secure': False, 'score': 0, 'details': {}}
    
    # Simple security check
    type_matches = [m for m in good_matches if m['template'].get('type') == consensus_type]
    score = len(type_matches) / len(good_matches) * 5  # Scale to 0-5
    
    return {
        'secure': score >= 3,
        'score': int(score),
        'details': {'type_fraction': len(type_matches) / len(good_matches)}
    }


# ----------------------------------------------------------------------
# 7.  Exports
# ----------------------------------------------------------------------
__all__ = [
    "SNIDResult",
    "compute_type_fractions", "compute_subtype_fractions",
    # Legacy functions (simplified)
    "determine_best_type", "determine_type_security_detailed",
    # Utility functions
    "get_main_type_from_template", "get_type_indices_from_template",
    "linear_fit_weighted",
    # Constants
    "RLAP_MIN", "MAXRLAP", "EPSRLAP", "EPSFRAC"
]


"""
Template Data Models
===================

Data models for template information and management.
"""

import logging
from typing import Dict, List, Optional, Any
import numpy as np

# Import logging
try:
    from snid_sage.shared.utils.logging import get_logger
    _LOGGER = get_logger('template_manager.data')
except ImportError:
    import logging
    _LOGGER = logging.getLogger('template_manager.data')

# H5PY for template storage access
try:
    import h5py
    H5PY_AVAILABLE = True
except ImportError:
    H5PY_AVAILABLE = False


class TemplateData:
    """Container for template data and metadata"""
    
    def __init__(self, name: str, template_info: Dict[str, Any]):
        self.name = name
        self.info = template_info
        self.flux_data = None
        self.wave_data = None
        self.epochs = []
        
    def load_data(self, storage_path: str):
        """Load template flux data from storage"""
        try:
            if not H5PY_AVAILABLE:
                raise ImportError("h5py not available for template data loading")
            
            _LOGGER.info(f"Loading template data for {self.name} from {storage_path}")
            
            # Load data from HDF5 storage file
            with h5py.File(storage_path, 'r') as f:
                _LOGGER.debug(f"HDF5 file keys: {list(f.keys())}")
                
                # Load wavelength grid from metadata
                if 'metadata' in f and 'standard_wavelength' in f['metadata']:
                    self.wave_data = f['metadata']['standard_wavelength'][:]
                    _LOGGER.info(f"Loaded wavelength data: {len(self.wave_data)} points")
                else:
                    _LOGGER.warning("No wavelength data found in metadata")
                
                # Check if template exists in templates group
                template_found = False
                if 'templates' in f and self.name in f['templates']:
                    template_group = f['templates'][self.name]
                    template_found = True
                    _LOGGER.info(f"Found template {self.name} in templates group")
                    _LOGGER.debug(f"Template group keys: {list(template_group.keys())}")
                    
                    # Load flux data
                    self.epochs = []
                    
                    # Check for multiple epochs first
                    if 'epochs' in template_group:
                        epochs_group = template_group['epochs']
                        _LOGGER.info(f"Found epochs group with {len(epochs_group)} epochs")
                        
                        for epoch_name in sorted(epochs_group.keys()):
                            epoch_group = epochs_group[epoch_name]
                            flux_data = epoch_group['flux'][:]
                            # Prefer HDF5 attr; default to NaN if missing/unparseable
                            try:
                                age_attr = epoch_group.attrs.get('age', np.nan)
                                age = float(age_attr)
                            except Exception:
                                age = float('nan')
                            epoch_info = {
                                'age': age,
                                'flux': flux_data
                            }
                            self.epochs.append(epoch_info)
                            _LOGGER.debug(f"Loaded epoch {epoch_name} with age {age} and {len(flux_data)} flux points")
                        
                        # Sort epochs by age (finite first, ascending), NaNs last
                        try:
                            self.epochs = sorted(
                                self.epochs,
                                key=lambda e: (0 if np.isfinite(e.get('age', np.nan)) else 1,
                                               float(e.get('age', np.inf)) if np.isfinite(e.get('age', np.nan)) else np.inf)
                            )
                        except Exception:
                            pass
                        # Set default flux to first (earliest) epoch
                        if self.epochs:
                            self.flux_data = self.epochs[0]['flux']
                        
                    elif 'flux' in template_group:
                        # Single epoch template (pre-epochs layout)
                        flux_data = template_group['flux'][:]
                        # Prefer age from HDF5 attrs; default to NaN if missing
                        try:
                            age_attr = template_group.attrs.get('age', np.nan)
                            age = float(age_attr)
                        except Exception:
                            age = float('nan')
                        # Phase stored as attr when present; otherwise Unknown
                        epoch_info = {
                            'age': age,
                            'flux': flux_data
                        }
                        self.epochs.append(epoch_info)
                        _LOGGER.info(f"Loaded single epoch with age {age} and {len(flux_data)} flux points")
                        
                        # Set default flux
                        self.flux_data = flux_data
                    else:
                        _LOGGER.warning(f"No flux data found for template {self.name}")
                
                if not template_found:
                    _LOGGER.warning(f"Template {self.name} not found in HDF5 file")
                    raise ValueError(f"Template {self.name} not found in {storage_path}")
                        
        except Exception as e:
            _LOGGER.error(f"Error loading template data for {self.name}: {e}")
            # Create mock data for testing
            _LOGGER.info("Creating mock data as fallback")
            self._create_mock_data()
    
    def _create_mock_data(self):
        """Create mock data for testing when real data is not available"""
        # Create a simple mock spectrum
        self.wave_data = np.linspace(3000, 9000, 1000)
        self.flux_data = np.exp(-0.5 * ((self.wave_data - 5000) / 1000)**2) + 0.1 * np.random.normal(0, 1, 1000)
        
        # Create mock epochs
        ages = [-10, 0, 10, 20, 30]
        self.epochs = []
        
        for age in ages:
            # Simulate spectral evolution
            flux_variation = 1.0 + 0.1 * np.sin(age / 10.0)
            mock_flux = self.flux_data * flux_variation + 0.05 * np.random.normal(0, 1, len(self.flux_data))
            
            epoch_info = {
                'age': age,
                'flux': mock_flux,
                'phase': 'Mock Phase'
            }
            self.epochs.append(epoch_info)
    
    def get_epoch_by_index(self, index: int) -> Optional[Dict[str, Any]]:
        """Get epoch data by index"""
        if 0 <= index < len(self.epochs):
            return self.epochs[index]
        return None
    
    def get_epoch_by_age(self, age: float, tolerance: float = 1.0) -> Optional[Dict[str, Any]]:
        """Get epoch data closest to specified age"""
        if not self.epochs:
            return None
        
        closest_epoch = None
        min_diff = float('inf')
        
        for epoch in self.epochs:
            diff = abs(epoch['age'] - age)
            if diff < min_diff and diff <= tolerance:
                min_diff = diff
                closest_epoch = epoch
        
        return closest_epoch
    
    def get_age_range(self) -> tuple:
        """Get the age range of available epochs"""
        if not self.epochs:
            return (float('nan'), float('nan'))
        
        # Consider only finite ages
        finite_ages = []
        for epoch in self.epochs:
            try:
                a = float(epoch.get('age', float('nan')))
                if np.isfinite(a):
                    finite_ages.append(a)
            except Exception:
                continue
        if not finite_ages:
            return (float('nan'), float('nan'))
        
        return (min(finite_ages), max(finite_ages))
    
    def get_flux_range(self) -> tuple:
        """Get the flux range across all epochs"""
        all_flux = []
        
        for epoch in self.epochs:
            if epoch['flux'] is not None:
                all_flux.extend(epoch['flux'])
        
        if not all_flux:
            return (0.0, 1.0)
        
        return (min(all_flux), max(all_flux))
    
    def normalize_epoch_flux(self, epoch_index: int) -> Optional[np.ndarray]:
        """Get normalized flux for a specific epoch"""
        epoch = self.get_epoch_by_index(epoch_index)
        if epoch and epoch['flux'] is not None:
            flux = epoch['flux']
            return flux / np.median(flux)
        return None
    
    def has_valid_data(self) -> bool:
        """Check if template has valid spectral data"""
        return (self.wave_data is not None and 
                len(self.epochs) > 0 and 
                any(epoch['flux'] is not None for epoch in self.epochs))
    
    def get_metadata_summary(self) -> Dict[str, Any]:
        """Get a summary of template metadata"""
        age_range = self.get_age_range()
        flux_range = self.get_flux_range()
        
        return {
            'name': self.name,
            'type': self.info.get('type', 'Unknown'),
            'subtype': self.info.get('subtype', 'Unknown'),
            'epochs_count': len(self.epochs),
            'age_range': age_range,
            'flux_range': flux_range,
            'wavelength_range': (
                float(np.min(self.wave_data)) if self.wave_data is not None else 0,
                float(np.max(self.wave_data)) if self.wave_data is not None else 0
            ),
            'redshift': self.info.get('redshift', 0.0),
            'has_valid_data': self.has_valid_data()
        }


"""
Utilities for extracting robust per-match fields used across SNID SAGE.

Currently includes:
- extract_redshift_sigma: robust extraction of per-match redshift uncertainty
  with sensible fallbacks when explicit uncertainties are missing.
"""

from __future__ import annotations

from typing import Any, Dict

import numpy as np

# Global scaling for redshift uncertainties; fixed to 1.0 (no configurability).
Z_K = 1.0


def extract_redshift_sigma(match: Dict[str, Any]) -> float:
    """
    Extract a per-match redshift uncertainty with robust fallbacks.

    Tries common keys first; if unavailable, derives a conservative
    proxy from the best similarity metric; final fallback is a small
    constant.
    """
    # Common alternative keys observed in data/products
    for key in ("redshift_uncertainty", "z_unc", "sigma_z", "z_err", "redshift_error"):
        try:
            val = match.get(key)
            if isinstance(val, (int, float)):
                v = float(val)
                if np.isfinite(v) and v > 0:
                    return v
        except Exception:
            # Continue probing other keys
            pass

    # Fallback: derive a loose proxy from similarity metric
    # Higher metric -> smaller sigma; keep a conservative upper bound
    try:
        from snid_sage.shared.utils.math_utils import get_best_metric_value  # lazy import

        m = float(get_best_metric_value(match))
        if np.isfinite(m) and m > 0:
            return float(Z_K * 0.02 / np.sqrt(1.0 + m))
    except Exception:
        pass

    # Final conservative fallback
    return 0.02 * Z_K


__all__ = ["extract_redshift_sigma"]





"""
Direct GMM Clustering for SNID SAGE
===================================

This module implements GMM clustering directly on redshift values for 
template matching analysis without any transformations.

Key features:
1. Direct GMM clustering on redshift values (no transformations)
2. Type-specific clustering with BIC-based model selection
3. Cluster selection and scoring (top-5 within-cluster)
4. Weighted subtype determination within winning clusters
5. Statistical confidence assessment for subtype classification

The clustering works directly with redshift values using the same approach
as the transformation_comparison_test.py reference implementation.
"""

import numpy as np
from typing import Dict, List, Any, Optional, Tuple, Callable
from sklearn.mixture import GaussianMixture
import logging
import time
from collections import defaultdict

_LOGGER = logging.getLogger(__name__)


# Note: find_winning_cluster_exact_match has been replaced by find_winning_cluster_top5_method
# The new method uses top-5 best metric values (prefer RLAP-CCC; fallback to RLAP) with penalties for small clusters





def calculate_joint_subtype_estimates_from_cluster(
    cluster_matches: List[Dict[str, Any]], 
    target_subtype: str
) -> Tuple[float, float, float, float, float, int, int]:
    """
    Calculate joint weighted redshift and age estimates for a specific subtype within a cluster.
    
    Parameters
    ----------
    cluster_matches : List[Dict]
        List of template match dictionaries from a cluster
    target_subtype : str
        The specific subtype to calculate estimates for (e.g., 'IIn', 'IIP')
        
    Returns
    -------
    Tuple[float, float, float, float, float, int, int]
        (weighted_redshift, weighted_age, redshift_error, age_error, 
         redshift_age_covariance, subtype_template_count, subtype_age_template_count)
    """
    if not cluster_matches or not target_subtype:
        return np.nan, np.nan, np.nan, np.nan, np.nan, 0, 0
    
    # Filter matches to only include the target subtype
    subtype_matches = []
    total_subtype_count = 0
    
    for match in cluster_matches:
        template = match.get('template', {})
        subtype = template.get('subtype', 'Unknown')
        if not subtype or subtype.strip() == '':
            subtype = 'Unknown'
        
        if subtype == target_subtype:
            total_subtype_count += 1
            age = template.get('age', 0.0)
            # Only include templates with both redshift and finite age for joint estimation
            # Note: Negative ages are acceptable (pre-maximum light)
            if 'redshift' in match and np.isfinite(age):
                subtype_matches.append(match)
    
    if not subtype_matches:
        _LOGGER.warning(f"No matches with valid (redshift, age) pairs found for subtype '{target_subtype}' in cluster")
        return np.nan, np.nan, np.nan, np.nan, np.nan, total_subtype_count, 0
    
    # Use weighted estimators directly for the subtype
    from snid_sage.shared.utils.math_utils import (
        get_best_metric_value,
        estimate_weighted_redshift,
        estimate_weighted_epoch,
        weighted_redshift_error,
        weighted_epoch_error,
    )

    redshifts_for_estimation = []
    redshift_errors_for_estimation = []
    metric_values_for_redshift = []
    ages_for_estimation = []
    metric_values_for_age = []
    age_redshift_errors_for_estimation = []

    for match in subtype_matches:
        if 'redshift' in match:
            template = match.get('template', {})
            age = template.get('age', 0.0)
            metric_val = get_best_metric_value(match)
            z = match.get('redshift')
            z_err = match.get('redshift_error', 0.0)
            if z is not None and np.isfinite(z) and z_err > 0:
                redshifts_for_estimation.append(z)
                redshift_errors_for_estimation.append(z_err)
                metric_values_for_redshift.append(metric_val)
            if np.isfinite(age) and z_err > 0:
                ages_for_estimation.append(age)
                metric_values_for_age.append(metric_val)
                age_redshift_errors_for_estimation.append(z_err)

    if redshifts_for_estimation:
        z_mean = estimate_weighted_redshift(
            redshifts_for_estimation,
            redshift_errors_for_estimation,
            metric_values_for_redshift
        )
        z_error = weighted_redshift_error(
            redshifts_for_estimation,
            redshift_errors_for_estimation,
            metric_values_for_redshift
        )
    else:
        _LOGGER.warning("No valid redshift data found in cluster matches")
        z_mean, z_error = np.nan, np.nan

    if ages_for_estimation and redshift_errors_for_estimation:
        t_mean = estimate_weighted_epoch(
            ages_for_estimation,
            age_redshift_errors_for_estimation,
            metric_values_for_age
        )
        t_error = weighted_epoch_error(
            ages_for_estimation,
            age_redshift_errors_for_estimation,
            metric_values_for_age
        )
    else:
        _LOGGER.warning("No valid age data found in cluster matches")
        t_mean, t_error = np.nan, np.nan

    zt_covariance = 0.0
    
    age_template_count = len(subtype_matches)
    
    return z_mean, t_mean, z_error, t_error, zt_covariance, total_subtype_count, age_template_count


def _compute_weighted_cluster_stats(
    cluster_matches: List[Dict[str, Any]]
) -> Tuple[float, float, float, float, float]:
    """Compute weighted redshift/age and their errors (unbiased weighted SD) for a set of matches.
    Returns (z_mean, t_mean, z_err, t_err, zt_covariance)."""
    from snid_sage.shared.utils.math_utils import (
        get_best_metric_value,
        estimate_weighted_redshift,
        estimate_weighted_epoch,
        weighted_redshift_error,
        weighted_epoch_error,
    )

    redshifts_for_estimation = []
    redshift_errors_for_estimation = []
    metric_values_for_redshift = []
    ages_for_estimation = []
    metric_values_for_age = []
    age_redshift_errors_for_estimation = []

    for match in cluster_matches:
        if 'redshift' in match:
            template = match.get('template', {})
            age = template.get('age', 0.0)
            metric_val = get_best_metric_value(match)
            z = match.get('redshift')
            z_err = match.get('redshift_error', 0.0)
            if z is not None and np.isfinite(z) and z_err > 0:
                redshifts_for_estimation.append(z)
                redshift_errors_for_estimation.append(z_err)
                metric_values_for_redshift.append(metric_val)
            if np.isfinite(age) and z_err > 0:
                ages_for_estimation.append(age)
                metric_values_for_age.append(metric_val)
                age_redshift_errors_for_estimation.append(z_err)

    if redshifts_for_estimation:
        z_mean = estimate_weighted_redshift(
            redshifts_for_estimation,
            redshift_errors_for_estimation,
            metric_values_for_redshift
        )
        z_err = weighted_redshift_error(
            redshifts_for_estimation,
            redshift_errors_for_estimation,
            metric_values_for_redshift
        )
    else:
        _LOGGER.warning("No valid redshift data found in cluster matches")
        z_mean, z_err = np.nan, np.nan

    if ages_for_estimation and redshift_errors_for_estimation:
        t_mean = estimate_weighted_epoch(
            ages_for_estimation,
            age_redshift_errors_for_estimation,
            metric_values_for_age
        )
        t_err = weighted_epoch_error(
            ages_for_estimation,
            age_redshift_errors_for_estimation,
            metric_values_for_age
        )
    else:
        _LOGGER.warning("No valid age data found in cluster matches")
        t_mean, t_err = np.nan, np.nan

    zt_covariance = 0.0

    return z_mean, t_mean, z_err, t_err, zt_covariance

    


def perform_direct_gmm_clustering(
    matches: List[Dict[str, Any]], 
    min_matches_per_type: int = 2,
    quality_threshold: float = 0.02,  # Direct redshift threshold
    max_clusters_per_type: int = 10,
    verbose: bool = False,
    rlap_ccc_threshold: float = 1.8,  # NEW: RLAP-CCC threshold for clustering
    use_weighted_gmm: Optional[bool] = None,  # Hidden option: when True, use weighted GMM + weighted BIC
    # Optional progress reporting (message, percent). If provided, emit per-type updates.
    progress_callback: Optional[Callable[[str, float], None]] = None,
    progress_min: float = 90.0,
    progress_max: float = 98.0
) -> Dict[str, Any]:
    """
    Direct GMM clustering on redshift values with automatic best metric selection.
    
    This approach works directly on redshift values without any transformations,
    matching the approach in transformation_comparison_test.py exactly.
    
    NEW: Now automatically uses the best available similarity metric via 
    get_best_metric_value(): prefers RLAP-CCC (final metric); falls back to RLAP when needed.
    
    Parameters
    ----------
    matches : List[Dict[str, Any]]
        List of template matches from SNID analysis
    min_matches_per_type : int, optional
        Minimum number of matches required per type for clustering
    quality_threshold : float, optional
        Redshift span threshold for cluster quality assessment
    max_clusters_per_type : int, optional
        Maximum clusters for GMM
    verbose : bool, optional
        Enable detailed logging
    rlap_ccc_threshold : float, optional
        Minimum RLAP-CCC value required for matches to be considered for clustering
        
    Hidden Options
    --------------
    use_weighted_gmm : bool, optional
        When True, enables sample-weighted GMM fitting and weighted-BIC selection
        using combined weights (metric^2 / sigma_z^2). By default (None/False),
        the method uses standard unweighted GMM and unweighted BIC. If None, the
        value is read from environment variable 'SNID_SAGE_WEIGHTED_GMM' (1/true/on).

    Returns
    -------
    Dict containing clustering results
    """
    
    start_time = time.time()
    
    # Resolve hidden option: default to unweighted unless explicitly enabled
    if use_weighted_gmm is None:
        try:
            import os
            env_val = str(os.getenv("SNID_SAGE_WEIGHTED_GMM", "0")).strip().lower()
            use_weighted_gmm = env_val in ("1", "true", "yes", "on")
        except Exception:
            use_weighted_gmm = False
    
    # Determine which metric to use - now using get_best_metric_value()
    # This automatically prioritizes RLAP-CCC > RLAP
    # Standardize metric naming; do not imply comparison phrasing.
    # We always prefer the final RLAP-CCC metric when available.
    metric_name = "RLAP-CCC"
    metric_key = "best_metric"  # Not actually used anymore, see get_best_metric_value() calls
    
    _LOGGER.info(f"🔄 Starting direct GMM {metric_name} clustering")
    _LOGGER.info(f"📐 Quality threshold: {quality_threshold:.3f} in redshift space")
    _LOGGER.info(f"🎯 RLAP-CCC threshold: {rlap_ccc_threshold:.1f} (matches below this are excluded from clustering)")
    
    # Filter matches by RLAP-CCC threshold before grouping
    from snid_sage.shared.utils.math_utils import get_best_metric_value
    filtered_matches = []
    excluded_count = 0
    
    for match in matches:
        metric_value = get_best_metric_value(match)
        if metric_value >= rlap_ccc_threshold:
            filtered_matches.append(match)
        else:
            excluded_count += 1
    
    if excluded_count > 0:
        _LOGGER.info(f"🙅 Filtered out {excluded_count} matches below RLAP-CCC threshold {rlap_ccc_threshold:.1f}")
        _LOGGER.info(f"✅ Proceeding with {len(filtered_matches)} matches for clustering")
    
    if not filtered_matches:
        _LOGGER.info(f"No matches above RLAP-CCC threshold {rlap_ccc_threshold:.1f}")
        return {'success': False, 'reason': 'no_matches_above_threshold'}
    
    # Group filtered matches by type
    type_groups = {}
    for match in filtered_matches:
        sn_type = match['template'].get('type', 'Unknown')
        if sn_type not in type_groups:
            type_groups[sn_type] = []
        type_groups[sn_type].append(match)
    
    # Accept all types with at least min_matches_per_type (now allowing 1+)
    filtered_type_groups = {
        sn_type: type_matches 
        for sn_type, type_matches in type_groups.items() 
        if len(type_matches) >= min_matches_per_type
    }
    
    if not filtered_type_groups:
        _LOGGER.info("No types have any matches for clustering")
        return {'success': False, 'reason': 'no_matches'}
    
    n_types = len(filtered_type_groups)
    _LOGGER.info(f"📊 Processing {n_types} types: {list(filtered_type_groups.keys())}")
    # Emit an initial clustering progress update if requested
    try:
        if progress_callback is not None and n_types > 0:
            progress_callback(
                f"Clustering: processing {n_types} type(s)",
                float(progress_min)
            )
    except Exception:
        pass
    
    # Perform GMM clustering for each type
    all_cluster_candidates = []
    clustering_results = {}
    
    # Iterate deterministically for stable UX
    for idx, sn_type in enumerate(sorted(filtered_type_groups.keys())):
        type_matches = filtered_type_groups[sn_type]
        type_result = _perform_direct_gmm_clustering(
            type_matches, sn_type, max_clusters_per_type, 
            quality_threshold, verbose, "best_metric",  # Parameter is deprecated
            use_weighted_gmm=use_weighted_gmm
        )
        
        clustering_results[sn_type] = type_result
        
        # Proceed when we have clusters, regardless of having a fitted GMM model
        if type_result.get('success') and type_result.get('clusters'):
            # For each cluster, use the EXACT same winning cluster selection as reference
            type_redshifts = np.array([m['redshift'] for m in type_matches])
            from snid_sage.shared.utils.math_utils import get_best_metric_value
            type_metric_values = np.array([get_best_metric_value(m) for m in type_matches])
            
            # Get cluster labels for this type (prefer GMM; fallback to gamma/argmax; else single cluster)
            features = type_redshifts.reshape(-1, 1)
            if 'gmm_model' in type_result and type_result['gmm_model'] is not None:
                labels = type_result['gmm_model'].predict(features)
            elif 'gamma' in type_result and isinstance(type_result['gamma'], np.ndarray):
                try:
                    labels = np.argmax(type_result['gamma'], axis=1)
                except Exception:
                    labels = np.zeros(len(type_matches), dtype=int)
            else:
                labels = np.zeros(len(type_matches), dtype=int)
            
            # Note: winning_cluster_id is now determined by the new top-5 method at the end
            # We don't need this old selection here anymore
            
            # Create cluster candidates using the exact reference approach
            for cluster_id in range(type_result['optimal_n_clusters']):
                cluster_info = next((c for c in type_result['clusters'] if c['id'] == cluster_id), None)
                if cluster_info is None:
                    continue
                
                # Calculate mean metric value for this cluster
                cluster_metric_values = [match.get(metric_key, match.get('rlap', 0)) for match in cluster_info['matches']]
                mean_metric = np.mean(cluster_metric_values) if cluster_metric_values else 0.0
                
                cluster_candidate = {
                    'type': sn_type,
                    'cluster_id': cluster_info['id'],
                    'matches': cluster_info['matches'],
                    'size': cluster_info['size'],
                    'mean_rlap': cluster_info['mean_rlap'],  # Keep original RLAP for compatibility
                    'mean_metric': mean_metric,  # NEW: Mean of selected metric (RLAP-CCC or RLAP)
                    'metric_name': metric_name,  # NEW: Name of metric used
                    'redshift_span': cluster_info['redshift_span'],
                    'redshift_quality': cluster_info['redshift_quality'],
                    'cluster_method': 'direct_gmm',
                    'quality_score': 0, # This will be updated by the new method
                    'composite_score': 0, # This will be updated by the new method
                    'is_winning_cluster': False  # Will be determined by new method
                    # Note: enhanced_redshift and other joint estimates will be added below
                }
                
                # Calculate subtype information for this cluster
                try:
                    # Get the type matches and gamma matrix for subtype calculation
                    type_data = clustering_results[sn_type]
                    if type_data.get('success') and 'gamma' in type_data:
                        gamma = type_data['gamma']
                        cluster_idx = cluster_info['id']  # cluster_id is the index within the type
                        
                        # Calculate subtype information for this specific cluster
                        best_subtype, subtype_confidence, subtype_margin_over_second, second_best_subtype = choose_subtype_weighted_voting(
                            sn_type, cluster_idx, type_matches, gamma
                        )
                        
                        # Calculate joint subtype-specific redshift and age estimates for the winning subtype
                        subtype_redshift = np.nan
                        subtype_redshift_err = np.nan
                        subtype_age = np.nan
                        subtype_age_err = np.nan
                        subtype_redshift_age_covariance = np.nan
                        subtype_template_count = 0
                        subtype_age_template_count = 0
                        
                        if best_subtype and best_subtype != 'Unknown':
                            # Use joint estimation for subtype
                            (subtype_redshift, subtype_age, subtype_redshift_err, subtype_age_err, 
                             _, subtype_template_count, subtype_age_template_count) = calculate_joint_subtype_estimates_from_cluster(
                                cluster_candidate['matches'], best_subtype
                            )
                        
                        # Calculate weighted estimates for the full cluster as well
                        cluster_redshift, cluster_age, cluster_redshift_err, cluster_age_err, _ = _compute_weighted_cluster_stats(
                            cluster_candidate['matches']
                        )
                        
                        # Add subtype information to cluster candidate
                        cluster_candidate.update({
                            'subtype_info': {
                                'best_subtype': best_subtype,
                                'subtype_confidence': subtype_confidence,
                                'subtype_margin_over_second': subtype_margin_over_second,
                                'second_best_subtype': second_best_subtype
                            },
                            # Add subtype-specific joint estimates
                            'subtype_redshift': subtype_redshift,
                            'subtype_redshift_err': subtype_redshift_err,
                            'subtype_age': subtype_age,
                            'subtype_age_err': subtype_age_err,
                            'subtype_template_count': subtype_template_count,
                            'subtype_age_template_count': subtype_age_template_count,
                            # Add full cluster joint estimates (update the enhanced_redshift from weighted_mean_redshift)
                            'enhanced_redshift': cluster_redshift,
                            'weighted_redshift_err': cluster_redshift_err,
                            'cluster_age': cluster_age,
                            'cluster_age_err': cluster_age_err
                        })
                        
                        if verbose:
                            _LOGGER.info(f"  Cluster {cluster_id} subtypes: {best_subtype} "
                                        f"(confidence: {subtype_confidence:.3f}, margin: {subtype_margin_over_second:.3f}, second: {second_best_subtype})")
                            
                            if not np.isnan(subtype_redshift) and not np.isnan(subtype_age):
                                _LOGGER.info(f"  Subtype {best_subtype} joint estimates: z={subtype_redshift:.6f}±{subtype_redshift_err:.6f}, "
                                           f"age={subtype_age:.1f}±{subtype_age_err:.1f} days "
                                           f"(from {subtype_age_template_count} templates with both redshift and age)")
                            elif not np.isnan(subtype_redshift):
                                _LOGGER.info(f"  Subtype {best_subtype} redshift: {subtype_redshift:.6f} ± {subtype_redshift_err:.6f}")
                                _LOGGER.warning(f"  Could not calculate joint age estimate for subtype {best_subtype}")
                            else:
                                _LOGGER.warning(f"  Could not calculate joint estimates for subtype {best_subtype}")
                            
                            if not np.isnan(cluster_redshift) and not np.isnan(cluster_age):
                                _LOGGER.info(f"  Full cluster joint estimates: z={cluster_redshift:.6f}±{cluster_redshift_err:.6f}, "
                                           f"age={cluster_age:.1f}±{cluster_age_err:.1f} days")
                except Exception as e:
                    # If subtype calculation fails, add default values
                    # Still calculate full cluster weighted estimates
                    try:
                        cluster_redshift, cluster_age, cluster_redshift_err, cluster_age_err, _ = _compute_weighted_cluster_stats(
                            cluster_candidate['matches']
                        )
                    except:
                        cluster_redshift = cluster_age = cluster_redshift_err = cluster_age_err = _ = np.nan
                    
                    cluster_candidate.update({
                        'subtype_info': {
                            'best_subtype': 'Unknown',
                            'subtype_confidence': 0.0,
                            'subtype_margin_over_second': 0.0,
                            'second_best_subtype': None
                        },
                        # Add default subtype values
                        'subtype_redshift': np.nan,
                        'subtype_redshift_err': np.nan,
                        'subtype_age': np.nan,
                        'subtype_age_err': np.nan,
                        'subtype_template_count': 0,
                        'subtype_age_template_count': 0,
                        # Add full cluster joint estimates (fallback)
                        'enhanced_redshift': cluster_redshift,
                        'weighted_redshift_err': cluster_redshift_err,
                        'cluster_age': cluster_age,
                        'cluster_age_err': cluster_age_err
                    })
                    if verbose:
                        _LOGGER.warning(f"  Failed to calculate subtypes for cluster {cluster_id}: {e}")
                
                all_cluster_candidates.append(cluster_candidate)

        # Per-type progress update
        try:
            if progress_callback is not None and n_types > 0:
                frac = (idx + 1) / float(n_types)
                pct = float(progress_min + (progress_max - progress_min) * frac)
                progress_callback(f"Clustering type {sn_type} ({idx + 1}/{n_types})", pct)
        except Exception:
            pass
    
    # Select best cluster using the new top-5 best metric method
    if not all_cluster_candidates:
        _LOGGER.info("No valid clusters found")
        return {'success': False, 'reason': 'no_clusters'}
    
    # Use the new top-5 method for cluster selection
    best_cluster, quality_assessment = find_winning_cluster_top5_method(
        all_cluster_candidates,
        verbose=verbose
    )
    
    if best_cluster is None:
        _LOGGER.warning("New cluster selection method failed")
        return {'success': False, 'reason': 'cluster_selection_failed'}
    
    # Update the best cluster with new quality metrics
    best_cluster['quality_assessment'] = quality_assessment['quality_assessment']
    best_cluster['confidence_assessment'] = quality_assessment['confidence_assessment']
    best_cluster['selection_method'] = 'top5_rlap_ccc'
    
    total_time = time.time() - start_time
    
    if verbose:
        _LOGGER.info("🏆 All cluster candidates (before new selection method):")
        for i, candidate in enumerate(all_cluster_candidates[:5]):
            _LOGGER.info(f"   {i+1}. {candidate['type']} cluster {candidate['cluster_id']}: "
                        f"size={candidate['size']}, z-span={candidate['redshift_span']:.4f}, "
                        f"quality={candidate['redshift_quality']}")
    
    _LOGGER.info(f"✅ Direct GMM clustering completed in {total_time:.3f}s")
    _LOGGER.info(f"Best cluster: {best_cluster['type']} cluster {best_cluster.get('cluster_id', 0)} "
                 f"(Quality: {best_cluster['quality_assessment']['quality_category']}, "
                 f"Confidence: {best_cluster['confidence_assessment']['confidence_level']})")
    
    return {
        'success': True,
        'method': 'direct_gmm',
        'metric_used': metric_name,  # NEW: Which metric was used
        'selection_method': 'top5_rlap_ccc',  # NEW: Selection method used
        'type_clustering_results': clustering_results,
        'best_cluster': best_cluster,
        'all_candidates': all_cluster_candidates,
        'quality_assessment': quality_assessment,  # NEW: Complete quality assessment
        'quality_threshold': quality_threshold,
        'total_computation_time': total_time,
        'n_types_clustered': len(clustering_results),
        'use_weighted_gmm': bool(use_weighted_gmm),
        'total_candidates': len(all_cluster_candidates)
    }


def _perform_direct_gmm_clustering(
    type_matches: List[Dict[str, Any]], 
    sn_type: str,
    max_clusters: int,
    quality_threshold: float,
    verbose: bool,
    metric_key: str = 'best_metric',  # DEPRECATED: Now uses get_best_metric_value() automatically
    *,
    use_weighted_gmm: bool = False
) -> Dict[str, Any]:
    """
    Perform GMM clustering directly on redshift values using the same approach
    as transformation_comparison_test.py.
    """
    
    try:
        redshifts = np.array([m['redshift'] for m in type_matches])
        rlaps = np.array([m['rlap'] for m in type_matches])  # Keep for compatibility
        from snid_sage.shared.utils.math_utils import get_best_metric_value, calculate_combined_weights
        from snid_sage.shared.utils.match_utils import extract_redshift_sigma
        metric_values = np.array([get_best_metric_value(m) for m in type_matches])  # Use best available metric (CCC > Cos > RLAP)
        sigmas = np.array([extract_redshift_sigma(m) for m in type_matches], dtype=float)
        
        # Suppress sklearn convergence warnings for cleaner output
        import warnings
        from sklearn.exceptions import ConvergenceWarning
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        
        # Step 1: Optionally build double weights and select model with weighted or unweighted BIC
        n_matches = len(type_matches)
        max_clusters_actual = min(max_clusters, n_matches // 2 + 1)
        
        if max_clusters_actual < 2:
            # Single match or too few for multi-cluster GMM - create single cluster
            _LOGGER.info(f"Creating single cluster for {sn_type} ({n_matches} matches)")
            return _create_single_cluster_result(
                type_matches, sn_type, redshifts, rlaps, quality_threshold, "best_metric"
            )
        
        # Build weights only when requested; default to unweighted GMM
        if use_weighted_gmm:
            # Double weights: (metric)^2 / sigma^2
            raw_weights = calculate_combined_weights(metric_values, sigmas)
            # Guard against degenerate sums
            sum_w = float(np.sum(raw_weights)) if raw_weights.size else 0.0
            if sum_w > 0:
                weights = raw_weights * (len(raw_weights) / sum_w)
            else:
                weights = np.ones_like(metric_values, dtype=float)
        else:
            weights = None

        bic_scores = []
        models = []
        
        for n_clusters in range(1, max_clusters_actual + 1):
            gmm = GaussianMixture(
                n_components=n_clusters, 
                random_state=42,
                max_iter=200,  # Same as transformation_comparison_test.py
                covariance_type='full',  # Same as transformation_comparison_test.py
                tol=1e-6  # Same as transformation_comparison_test.py
            )

            # Cluster directly on redshift values (no transformation)
            features = redshifts.reshape(-1, 1)

            if use_weighted_gmm and weights is not None:
                # Weighted fit and weighted BIC; fallback to resampling if needed
                d = features.shape[1]
                try:
                    gmm.fit(features, sample_weight=weights)
                    logprob = gmm.score_samples(features)
                    # Parameter count for full covariance
                    p = (n_clusters - 1) + n_clusters * d + n_clusters * d * (d + 1) / 2.0
                    bic = -2.0 * float(np.sum(weights * logprob)) + float(p) * np.log(float(np.sum(weights)))
                except TypeError:
                    # Resampling fallback
                    rng = np.random.RandomState(42)
                    # Target size similar to test script bounds
                    target = int(min(max(len(features) * 5, 300), 5000)) if len(features) > 0 else 0
                    if target > 0:
                        p_norm = weights / float(np.sum(weights)) if np.sum(weights) > 0 else np.full_like(weights, 1.0 / len(weights))
                        idx = rng.choice(np.arange(len(features)), size=target, replace=True, p=p_norm)
                        features_rs = features[idx]
                        gmm.fit(features_rs)
                        bic = float(gmm.bic(features_rs))
                    else:
                        gmm.fit(features)
                        bic = float(gmm.bic(features))
            else:
                # Unweighted default path
                gmm.fit(features)
                bic = float(gmm.bic(features))

            bic_scores.append(bic)
            models.append(gmm)
        
        # Select optimal model (minimum BIC)
        optimal_idx = np.argmin(bic_scores)
        optimal_n_clusters = optimal_idx + 1
        best_gmm = models[optimal_idx]

        # Get cluster assignments and responsibilities
        features = redshifts.reshape(-1, 1)
        labels = best_gmm.predict(features)
        gamma = best_gmm.predict_proba(features)

        # Enforce contiguity in 1D redshift: split any non-contiguous cluster into
        # contiguous segments along sorted redshift order, then split by hard gaps
        order = np.argsort(redshifts)
        labels_sorted = labels[order]

        # Run-length encode labels along sorted z
        runs = []  # list of (label, start_idx_in_sorted, end_idx_in_sorted)
        start = 0
        for i in range(1, len(labels_sorted) + 1):
            if i == len(labels_sorted) or labels_sorted[i] != labels_sorted[i - 1]:
                runs.append((labels_sorted[i - 1], start, i))
                start = i

        # Build contiguous segments per original label
        segments = []  # list of (orig_label, absolute_indices)
        label_to_run_count = {k: 0 for k in range(optimal_n_clusters)}
        min_segment_size = 1  # keep even tiny side clusters; scoring penalizes later
        for orig_label in range(optimal_n_clusters):
            run_spans = [(s, e) for lbl, s, e in runs if lbl == orig_label]
            label_to_run_count[orig_label] = len(run_spans)
            for (s, e) in run_spans:
                idx = order[s:e]
                if len(idx) >= min_segment_size:
                    segments.append((orig_label, idx))

        split_applied = any(cnt > 1 for cnt in label_to_run_count.values())

        final_clusters = []
        # Hard gap split: 0.025
        MAX_GAP_Z = 0.025

        # Helper: split a sorted-by-z absolute index run by redshift gaps
        def _split_by_gap(abs_idx: np.ndarray, z: np.ndarray) -> List[np.ndarray]:
            if abs_idx.size <= 1:
                return [abs_idx]
            parts: List[np.ndarray] = []
            start = 0
            for r in range(1, abs_idx.size):
                if abs(z[abs_idx[r]] - z[abs_idx[r-1]]) > MAX_GAP_Z:
                    parts.append(abs_idx[start:r])
                    start = r
            parts.append(abs_idx[start:])
            return parts

        # Build final segments (run-contiguity plus gap splits)
        segment_records: List[Tuple[int, np.ndarray, bool]] = []  # (orig_label, indices, is_gap_split)
        for orig_label, idx in segments:
            # idx is in absolute index order of the run; ensure it is sorted by z for consistent gap checks
            idx_sorted = idx[np.argsort(redshifts[idx])]
            parts = _split_by_gap(idx_sorted, redshifts)
            if len(parts) <= 1:
                segment_records.append((orig_label, idx_sorted, False))
            else:
                for j, part in enumerate(parts):
                    segment_records.append((orig_label, part, True))

        if segment_records:
            # Rebuild responsibilities with one column per final segment
            n_segments = len(segment_records)
            new_gamma = np.zeros((len(type_matches), n_segments), dtype=float)
            for j, (orig_label, idx, _is_gap) in enumerate(segment_records):
                new_gamma[idx, j] = gamma[idx, orig_label]
            gamma = new_gamma

            # Build clusters from final segments
            for new_id, (orig_label, idx, is_gap) in enumerate(segment_records):
                cluster_redshifts = redshifts[idx]
                cluster_rlaps = rlaps[idx]
                cluster_metric_values = metric_values[idx]
                cluster_matches = [type_matches[i] for i in idx]

                redshift_span = float(np.max(cluster_redshifts) - np.min(cluster_redshifts)) if len(cluster_redshifts) > 0 else 0.0
                if redshift_span <= quality_threshold:
                    redshift_quality = 'tight'
                elif redshift_span <= quality_threshold * 2:
                    redshift_quality = 'moderate'
                elif redshift_span <= quality_threshold * 4:
                    redshift_quality = 'loose'
                else:
                    redshift_quality = 'very_loose'

                weighted_mean_redshift, _, weighted_redshift_err, _, _ = _compute_weighted_cluster_stats(cluster_matches)

                final_clusters.append({
                    'id': new_id,
                    'matches': cluster_matches,
                    'size': len(cluster_matches),
                    'mean_rlap': float(np.mean(cluster_rlaps)) if len(cluster_rlaps) > 0 else 0.0,
                    'std_rlap': float(np.std(cluster_rlaps)) if len(cluster_rlaps) > 1 else 0.0,
                    'mean_metric': float(np.mean(cluster_metric_values)) if len(cluster_metric_values) > 0 else 0.0,
                    'std_metric': float(np.std(cluster_metric_values)) if len(cluster_metric_values) > 1 else 0.0,
                    'metric_key': metric_key,
                    'weighted_mean_redshift': float(weighted_mean_redshift) if np.isfinite(weighted_mean_redshift) else np.nan,
                    'weighted_redshift_err': float(weighted_redshift_err) if np.isfinite(weighted_redshift_err) else np.nan,
                    'redshift_span': redshift_span,
                    'redshift_quality': redshift_quality,
                    'cluster_method': 'direct_gmm_contiguous',
                    'rlap_range': (float(np.min(cluster_rlaps)), float(np.max(cluster_rlaps))) if len(cluster_rlaps) > 0 else (0.0, 0.0),
                    'metric_range': (float(np.min(cluster_metric_values)), float(np.max(cluster_metric_values))) if len(cluster_metric_values) > 0 else (0.0, 0.0),
                    'redshift_range': (float(np.min(cluster_redshifts)), float(np.max(cluster_redshifts))) if len(cluster_redshifts) > 0 else (0.0, 0.0),
                    'top_5_values': [],
                    'top_5_mean': 0.0,
                    'penalty_factor': 1.0,
                    'penalized_score': 0.0,
                    'composite_score': 0.0,
                    # New annotations used by tests/plots
                    'segment_id': new_id,
                    'gap_split': bool(is_gap),
                    'indices': [int(v) for v in idx.tolist()],
                })

                if verbose:
                    _LOGGER.info(f"  Segment {new_id} (from label {orig_label}): {redshift_quality} (z-span={redshift_span:.4f})")

            # Replace optimal cluster count with the number of final segments
            optimal_n_clusters = len(final_clusters)
        else:
            # Create cluster info from original labels (already contiguous)
            for cluster_id in range(optimal_n_clusters):
                cluster_mask = (labels == cluster_id)
                cluster_indices = np.where(cluster_mask)[0]

                if len(cluster_indices) < 1:
                    continue

                cluster_redshifts = redshifts[cluster_mask]
                cluster_rlaps = rlaps[cluster_mask]
                cluster_metric_values = metric_values[cluster_mask]
                cluster_matches = [type_matches[i] for i in cluster_indices]

                redshift_span = np.max(cluster_redshifts) - np.min(cluster_redshifts)
                if redshift_span <= quality_threshold:
                    redshift_quality = 'tight'
                elif redshift_span <= quality_threshold * 2:
                    redshift_quality = 'moderate'
                elif redshift_span <= quality_threshold * 4:
                    redshift_quality = 'loose'
                else:
                    redshift_quality = 'very_loose'

                weighted_mean_redshift, _, weighted_redshift_err, _, _ = _compute_weighted_cluster_stats(cluster_matches)

                cluster_info = {
                    'id': cluster_id,
                    'matches': cluster_matches,
                    'size': len(cluster_matches),
                    'mean_rlap': np.mean(cluster_rlaps),
                    'std_rlap': np.std(cluster_rlaps) if len(cluster_rlaps) > 1 else 0.0,
                    'mean_metric': np.mean(cluster_metric_values),
                    'std_metric': np.std(cluster_metric_values) if len(cluster_metric_values) > 1 else 0.0,
                    'metric_key': metric_key,
                    'weighted_mean_redshift': weighted_mean_redshift,
                    'weighted_redshift_err': weighted_redshift_err,
                    'redshift_span': redshift_span,
                    'redshift_quality': redshift_quality,
                    'cluster_method': 'direct_gmm',
                    'rlap_range': (np.min(cluster_rlaps), np.max(cluster_rlaps)),
                    'metric_range': (np.min(cluster_metric_values), np.max(cluster_metric_values)),
                    'redshift_range': (np.min(cluster_redshifts), np.max(cluster_redshifts)),
                    'top_5_values': [],
                    'top_5_mean': 0.0,
                    'penalty_factor': 1.0,
                    'penalized_score': 0.0,
                    'composite_score': 0.0,
                    # Consistent annotations for UI/tests even when no gap split applied
                    'segment_id': cluster_id,
                    'gap_split': False,
                    'indices': [int(v) for v in cluster_indices.tolist()],
                }
                final_clusters.append(cluster_info)

                if verbose:
                    _LOGGER.info(f"  Cluster {cluster_id}: {redshift_quality} (z-span={redshift_span:.4f})")

        return {
            'success': True,
            'type': sn_type,
            'optimal_n_clusters': optimal_n_clusters,
            'final_n_clusters': len(final_clusters),
            'bic_scores': bic_scores,
            'clusters': final_clusters,
            'gmm_model': best_gmm,
            'gamma': gamma,
            'type_matches': type_matches,  # Store the original matches used for gamma matrix
            'quality_threshold': quality_threshold,
            'contiguity_split_applied': True if final_clusters else bool(split_applied),
            # Debug extras
            'weights': weights.tolist() if isinstance(weights, np.ndarray) else []
        }
                
    except Exception as e:
        _LOGGER.error(f"Direct GMM clustering failed for type {sn_type}: {e}")
        return {'success': False, 'type': sn_type, 'error': str(e)}


def _create_single_cluster_result(
    type_matches: List[Dict[str, Any]], 
    sn_type: str, 
    redshifts: np.ndarray, 
    rlaps: np.ndarray,
    quality_threshold: float,
    metric_key: str = 'best_metric'  # DEPRECATED: Now uses get_best_metric_value()
) -> Dict[str, Any]:
    """Create a single cluster result when clustering isn't possible/needed."""
    
    redshift_span = np.max(redshifts) - np.min(redshifts) if len(redshifts) > 1 else 0.0
    
    # Get metric values using best available metric
    from snid_sage.shared.utils.math_utils import get_best_metric_value
    metric_values = np.array([get_best_metric_value(m) for m in type_matches])
    
    # Quality based on redshift span
    if redshift_span <= quality_threshold:
        redshift_quality = 'tight'
    elif redshift_span <= quality_threshold * 2:
        redshift_quality = 'moderate'
    else:
        redshift_quality = 'loose'
    
    # Calculate enhanced redshift statistics using joint estimation (just extract redshift)
    weighted_mean_redshift, _, weighted_redshift_sd, _, _ = _compute_weighted_cluster_stats(type_matches)
    
    cluster_info = {
        'id': 0,
        'matches': type_matches,
        'size': len(type_matches),
        'mean_rlap': np.mean(rlaps),
        'std_rlap': np.std(rlaps) if len(rlaps) > 1 else 0.0,
        'mean_metric': np.mean(metric_values),  # NEW: Mean of selected metric
        'std_metric': np.std(metric_values) if len(metric_values) > 1 else 0.0,  # NEW
        'metric_key': metric_key,  # NEW: Which metric was used
                    # Enhanced redshift statistics
        'weighted_mean_redshift': weighted_mean_redshift,
        'weighted_redshift_err': weighted_redshift_sd,
        'redshift_span': redshift_span,
        'redshift_quality': redshift_quality,
        'cluster_method': 'single_cluster',
        'rlap_range': (np.min(rlaps), np.max(rlaps)),
        'metric_range': (np.min(metric_values), np.max(metric_values)),  # NEW
        'redshift_range': (np.min(redshifts), np.max(redshifts)),
        'top_5_values': [],
        'top_5_mean': 0.0,
        'penalty_factor': 1.0,
        'penalized_score': 0.0,
        'composite_score': 0.0
    }
    
    # Create a trivial responsibilities matrix (all weight to the single cluster)
    try:
        gamma = np.ones((len(type_matches), 1), dtype=float)
    except Exception:
        gamma = None

    return {
        'success': True,
        'type': sn_type,
        'optimal_n_clusters': 1,
        'final_n_clusters': 1,
        'clusters': [cluster_info],
        'quality_threshold': quality_threshold,
        'gamma': gamma,
        'type_matches': type_matches
    }


def choose_subtype_weighted_voting(
    winning_type: str, 
    k_star: int, 
    matches: List[Dict[str, Any]], 
    gamma: np.ndarray, 
    resp_cut: float = 0.001
) -> tuple:
    """
    Choose the best subtype within the winning cluster using top-5 best metric method.
    
    Args:
        winning_type: The winning type (e.g., "Ia")
        k_star: Index of the winning cluster within that type
        matches: List of template matches for the winning type
        gamma: GMM responsibilities matrix, shape (n_matches, n_clusters)
        resp_cut: Minimum responsibility threshold
    
    Returns:
        tuple: (best_subtype, confidence, margin_over_second, second_best_subtype)
    """
    
    # Collect cluster members
    cluster_members = []
    
    # Safety check: ensure gamma matrix dimensions match matches list
    if len(matches) != gamma.shape[0]:
        _LOGGER.error(f"Dimension mismatch: matches={len(matches)}, gamma.shape={gamma.shape}")
        raise ValueError(f"Matches list length ({len(matches)}) does not match gamma matrix rows ({gamma.shape[0]})")
    
    # Safety check: ensure k_star is valid cluster index
    if k_star >= gamma.shape[1]:
        _LOGGER.error(f"Cluster index out of bounds: k_star={k_star}, gamma.shape[1]={gamma.shape[1]}")
        raise ValueError(f"Cluster index {k_star} is out of bounds for gamma matrix with {gamma.shape[1]} clusters")
    
    for i, match in enumerate(matches):
        # Include members assigned to the selected cluster (no responsibility threshold)
        # Assignment is by argmax of responsibilities, matching how cluster members are defined downstream
        try:
            assigned_cluster = int(np.argmax(gamma[i, :]))
        except Exception:
            assigned_cluster = -1
        if assigned_cluster != k_star:
            continue
        
        subtype = match['template'].get('subtype', 'Unknown')
        if not subtype or subtype.strip() == '':
            subtype = 'Unknown'
        
        # Use best available metric (RLAP-CCC if available, otherwise RLAP)
        from snid_sage.shared.utils.math_utils import get_best_metric_value
        metric_value = get_best_metric_value(match)
        # Pull per-match redshift uncertainty if available
        sigma_z = match.get('redshift_error', match.get('z_err', match.get('sigma_z', None)))
        
        cluster_members.append({
            'subtype': subtype,
            'metric_value': metric_value,
            'cluster_membership': gamma[i, k_star],
            'redshift_error': sigma_z
        })
    
    if not cluster_members:
        return "Unknown", 0.0, 0.0, None
    
    # Group by subtype and calculate top-5 mean best metric for each
    subtype_groups = defaultdict(list)
    for member in cluster_members:
        subtype_groups[member['subtype']].append(member)
    
    # Calculate top-5 weighted mean for each subtype using weights w = metric^2 / sigma_z^2
    subtype_scores = {}
    for subtype, members in subtype_groups.items():
        # Sort by metric value (best available) descending
        sorted_members = sorted(members, key=lambda x: x['metric_value'], reverse=True)
        
        # Take top 5 (or all if less than 5)
        top_members = sorted_members[:5]
        top_values = [m['metric_value'] for m in top_members]
        # Compute weights using the same scheme as weighted redshift: w = (metric)^2 / sigma_z^2
        from snid_sage.shared.utils.math_utils import compute_cluster_weights
        metrics_array = np.asarray([m['metric_value'] for m in top_members], dtype=float)
        sigmas_array = np.asarray([
            (m.get('redshift_error') if m.get('redshift_error') is not None else np.nan)
            for m in top_members
        ], dtype=float)
        # Valid where finite and sigma > 0
        valid_mask = (np.isfinite(metrics_array) & np.isfinite(sigmas_array) & (sigmas_array > 0))
        if np.any(valid_mask):
            weights = compute_cluster_weights(metrics_array[valid_mask], sigmas_array[valid_mask])
            sum_w = np.sum(weights)
            mean_top = float(np.sum(weights * metrics_array[valid_mask]) / sum_w) if sum_w > 0 else (sum(top_values) / len(top_values))
        else:
            # Fallback to unweighted mean if no valid uncertainties
            mean_top = sum(top_values) / len(top_values)
        
        # Apply penalty if less than 5 templates
        penalty_factor = len(top_values) / 5.0  # 1.0 if 5 templates, 0.8 if 4, etc.
        
        # Final score = mean_top × penalty_factor
        subtype_scores[subtype] = mean_top * penalty_factor
    
    if not subtype_scores:
        return "Unknown", 0.0, 0.0, None
    
    # Find best subtype
    best_subtype = max(subtype_scores, key=subtype_scores.get)
    best_score = subtype_scores[best_subtype]
    
    # Calculate margin over second best
    sorted_scores = sorted(subtype_scores.values(), reverse=True)
    margin_over_second = sorted_scores[0] - (sorted_scores[1] if len(sorted_scores) > 1 else 0)
    
    # Convert score to confidence (0-1 range)
    total_score = sum(subtype_scores.values())
    confidence = best_score / total_score if total_score > 0 else 0.0
    
    # Calculate relative margin as percentage (more intuitive for display)
    relative_margin_pct = 0.0
    if len(sorted_scores) > 1 and sorted_scores[1] > 0:
        second_best_score = sorted_scores[1]
        relative_margin_pct = (margin_over_second / second_best_score) * 100
    
    # Get second best subtype if available
    second_best_subtype = None
    if len(sorted_scores) > 1:
        second_best_score = sorted_scores[1]
        # Find which subtype has this score
        for subtype, score in subtype_scores.items():
            if abs(score - second_best_score) < 1e-6:  # Float comparison
                second_best_subtype = subtype
                break
    
    return best_subtype, confidence, relative_margin_pct, second_best_subtype





def create_3d_visualization_data(clustering_results: Dict[str, Any]) -> Dict[str, np.ndarray]:
    """Prepare data for 3D visualization: redshift vs type vs RLAP/RLAP-CCC."""
    
    redshifts = []
    metric_values = []
    types = []
    type_indices = []
    cluster_ids = []
    matches = []  # Store matches for access to best metric values
    
    type_to_index = {}
    current_type_index = 0
    
    # Check if we have the new clustering structure with all_candidates
    if 'all_candidates' in clustering_results:
        # New structure: use all_candidates
        for candidate in clustering_results.get('all_candidates', []):
            sn_type = candidate.get('type', 'Unknown')
            if sn_type not in type_to_index:
                type_to_index[sn_type] = current_type_index
                current_type_index += 1
            
            type_index = type_to_index[sn_type]
            cluster_id = candidate.get('cluster_id', 0)
            
            for match in candidate.get('matches', []):
                redshifts.append(match['redshift'])
                # Use best available metric (RLAP-CCC if available, otherwise RLAP)
                from snid_sage.shared.utils.math_utils import get_best_metric_value
                metric_values.append(get_best_metric_value(match))
                types.append(sn_type)
                type_indices.append(type_index)
                cluster_ids.append(cluster_id)
                matches.append(match)
    
    else:
        # Fallback: old structure with type_clustering_results
        for type_result in clustering_results.get('type_clustering_results', {}).values():
            if not type_result.get('success', False):
                continue
                
            sn_type = type_result['type']
            if sn_type not in type_to_index:
                type_to_index[sn_type] = current_type_index
                current_type_index += 1
            
            type_index = type_to_index[sn_type]
            
            for cluster in type_result['clusters']:
                for match in cluster['matches']:
                    redshifts.append(match['redshift'])
                    # Use best available metric (RLAP-CCC if available, otherwise RLAP)
                    from snid_sage.shared.utils.math_utils import get_best_metric_value
                    metric_values.append(get_best_metric_value(match))
                    types.append(sn_type)
                    type_indices.append(type_index)
                    cluster_ids.append(cluster['id'])
                    matches.append(match)
    
    return {
        'redshifts': np.array(redshifts),
        'rlaps': np.array(metric_values),  # Keep key name for backward compatibility
        'types': types,
        'type_indices': np.array(type_indices),
        'cluster_ids': np.array(cluster_ids),
        'type_mapping': type_to_index,
        'matches': matches  # NEW: Include matches for access to best metric values
    }






def find_winning_cluster_top5_method(
    all_cluster_candidates: List[Dict[str, Any]], 
    verbose: bool = False
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Find the winning cluster using the top-5 best metric method (prefer RLAP-CCC; fallback to RLAP).
    
    This method:
    1. Takes the top 5 best metric values from each cluster (using get_best_metric_value())
    2. Calculates the mean of these top 5 values
    3. Penalizes clusters with fewer than 5 points
    4. Selects the cluster with the highest mean
    5. Provides confidence assessment vs other clusters
    6. Provides absolute quality assessment (Low/Mid/High)
    
    Parameters
    ----------
    all_cluster_candidates : List[Dict[str, Any]]
        List of all cluster candidates from GMM clustering
    verbose : bool, optional
        Enable detailed logging
        
    Returns
    -------
    Tuple[Dict[str, Any], Dict[str, Any]]
        (winning_cluster, quality_assessment)
    """
    if not all_cluster_candidates:
        return None, {'error': 'No cluster candidates available'}
    
    # DEPRECATED: Parameters not used anymore, see get_best_metric_value() calls
    # Standardize metric naming; do not imply comparison phrasing.
    # We always prefer the final RLAP-CCC metric when available.
    metric_name = 'RLAP-CCC'
    
    # Calculate top-5 means for each cluster
    cluster_scores = []
    
    for cluster in all_cluster_candidates:
        matches = cluster.get('matches', [])
        if not matches:
            continue
            
        # Extract (metric, sigma_z) pairs and sort descending by metric
        metric_sigma_pairs = []
        for match in matches:
            # Use get_best_metric_value to automatically prioritize RLAP-CCC
            from snid_sage.shared.utils.math_utils import get_best_metric_value
            metric = get_best_metric_value(match)
            sigma_z = match.get('redshift_error', match.get('z_err', match.get('sigma_z', None)))
            metric_sigma_pairs.append((float(metric), (float(sigma_z) if sigma_z is not None else np.nan)))
        
        metric_sigma_pairs.sort(key=lambda x: x[0], reverse=True)  # Highest metric first
        
        # Take top 5 (or all if fewer than 5)
        top_5_pairs = metric_sigma_pairs[:5]
        top_5_values = [m for m, _ in top_5_pairs]
        sigmas = [s for _, s in top_5_pairs]
        
        # Calculate weighted mean of top 5 using w = metric^2 / sigma_z^2; fallback to unweighted mean
        top_1_share = 0.0
        if top_5_values:
            from snid_sage.shared.utils.math_utils import compute_cluster_weights
            metrics_array = np.asarray(top_5_values, dtype=float)
            sigmas_array = np.asarray(sigmas, dtype=float)
            valid_mask = (np.isfinite(metrics_array) & np.isfinite(sigmas_array) & (sigmas_array > 0))
            if np.any(valid_mask):
                weights = compute_cluster_weights(metrics_array[valid_mask], sigmas_array[valid_mask])
                sum_w = np.sum(weights)
                top_5_mean = float(np.sum(weights * metrics_array[valid_mask]) / sum_w) if sum_w > 0 else float(np.mean(metrics_array))
                # Contribution share of the first/top match among the top-5 weights
                if sum_w > 0 and valid_mask.size > 0 and valid_mask[0]:
                    # Index 0 corresponds to the highest-metric match
                    # Map index 0 in full array to its position in the compressed valid array
                    pos0 = int(np.flatnonzero(valid_mask)[0])
                    top_1_share = float(weights[pos0] / sum_w) if pos0 < len(weights) else 0.0
                else:
                    top_1_share = 0.0
            else:
                top_5_mean = float(np.mean(metrics_array))
                top_1_share = 1.0 / float(len(metrics_array))
        else:
            top_5_mean = 0.0
            top_1_share = 0.0
        
        # Apply penalty using the same linear scheme as subtype selection: count/5.0 (capped at 1.0)
        penalty_factor = (len(top_5_values) / 5.0) if top_5_values else 0.0
        if penalty_factor > 1.0:
            penalty_factor = 1.0
        
        penalized_score = top_5_mean * penalty_factor  # Keep ALL clusters; penalize small ones linearly
        
        # Annotate the original cluster dictionary so downstream UIs can display these metrics
        cluster['top_5_values'] = top_5_values
        cluster['top_5_mean'] = top_5_mean
        cluster['penalty_factor'] = penalty_factor
        cluster['penalized_score'] = penalized_score
        # For convenience update composite_score field used in various summaries
        cluster['composite_score'] = penalized_score
        # Store the top-1 contribution share for diagnostics/reporting
        cluster['top_1_share'] = top_1_share
        
        cluster_info = {
            'cluster': cluster,
            'cluster_size': len(matches),
            'top_5_values': top_5_values,
            'top_5_mean': top_5_mean,
            'penalty_factor': penalty_factor,
            'penalized_score': penalized_score,
            'cluster_type': cluster.get('type', 'Unknown'),
            'cluster_id': cluster.get('cluster_id', 0),
            'top_1_share': top_1_share
        }
        
        cluster_scores.append(cluster_info)
    
    if not cluster_scores:
        return None, {'error': 'No valid clusters found'}
    
    # Sort by penalized score (highest first)
    cluster_scores.sort(key=lambda x: x['penalized_score'], reverse=True)
    
    # Winner is the cluster with highest penalized score
    winning_cluster_info = cluster_scores[0]
    winning_cluster = winning_cluster_info['cluster']
    
    # Calculate confidence assessment
    confidence_assessment = _calculate_cluster_confidence(cluster_scores, metric_name)
    
    # Calculate absolute quality assessment
    quality_assessment = _calculate_absolute_quality(winning_cluster_info, metric_name)
    
    # Combine assessments
    full_assessment = {
        'winning_cluster': winning_cluster,
        'winning_cluster_info': winning_cluster_info,
        'all_cluster_scores': cluster_scores,
        'confidence_assessment': confidence_assessment,
        'quality_assessment': quality_assessment,
        'metric_used': metric_name,
        'selection_method': 'top5_rlap_ccc'
    }
    
    if verbose:
        _log_cluster_selection_details(full_assessment)
    
    return winning_cluster, full_assessment


def _calculate_cluster_confidence(cluster_scores: List[Dict[str, Any]], metric_name: str) -> Dict[str, Any]:
    """Calculate relative confidence as percent improvement vs best different-type competitor.

    Returns a minimal assessment with:
      - confidence_pct: percent improvement (NaN if no different-type competitor)
      - confidence_level: discrete flag derived from confidence_pct
      - second_best_type: type of best different-type competitor (or 'N/A')
      - confidence_description: concise human-readable summary
    """
    # Not enough clusters to compare
    if len(cluster_scores) < 2:
        return {
            'confidence_pct': float('nan'),
            'confidence_level': 'No Comp',
            'confidence_description': 'Single cluster',
            'second_best_type': 'N/A'
        }

    # Find the first runner-up of a different type than the winner
    winning_type = cluster_scores[0]['cluster_type']
    competitor_info = None
    for i in range(1, len(cluster_scores)):
        if cluster_scores[i]['cluster_type'] != winning_type:
            competitor_info = cluster_scores[i]
            break

    # No different-type competitor -> undefined percent (NaN)
    if competitor_info is None:
        return {
            'confidence_pct': float('nan'),
            'confidence_level': 'No Comp',
            'confidence_description': 'No competitor',
            'second_best_type': 'N/A'
        }

    best_score = float(cluster_scores[0]['penalized_score'])
    second_best_score = float(competitor_info['penalized_score'])

    # Percent improvement; guard zero competitor
    if second_best_score > 0.0:
        confidence_pct = (best_score - second_best_score) / second_best_score * 100.0
    else:
        # Treat as undefined; display can clamp if desired
        confidence_pct = float('nan')

    # Discrete flag from percent thresholds
    if isinstance(confidence_pct, float) and not np.isnan(confidence_pct):
        # Updated thresholds: High ≥ 100%, Medium ≥ 30%, Low ≥ 10%, else Very Low
        if confidence_pct >= 100.0:
            confidence_level = 'High'
        elif confidence_pct >= 30.0:
            confidence_level = 'Medium'
        elif confidence_pct >= 10.0:
            confidence_level = 'Low'
        else:
            confidence_level = 'Very Low'
        confidence_description = f"{confidence_pct:.1f}% better than second best"
    else:
        confidence_level = 'N/A'
        confidence_description = 'N/A'

    return {
        'confidence_pct': confidence_pct,
        'confidence_level': confidence_level,
        'confidence_description': confidence_description,
        'second_best_type': competitor_info['cluster_type']
    }


def _calculate_absolute_quality(winning_cluster_info: Dict[str, Any], metric_name: str) -> Dict[str, Any]:
    """Calculate absolute quality assessment for the winning cluster using penalized top-5 score."""
    
    # Use the already calculated penalized score from the winning cluster
    penalized_score = winning_cluster_info['penalized_score']
    top_5_mean = winning_cluster_info['top_5_mean']
    penalty_factor = winning_cluster_info['penalty_factor']
    cluster_size = winning_cluster_info['cluster_size']
    
    # Quality categories based solely on penalized top-5 mean (global rule)
    #  - Very Low: < 2.5
    #  - Low: 2.5 to < 4.0
    #  - Medium: 4.0 to ≤ 10.0
    #  - High: > 10.0
    if penalized_score > 10.0:
        quality_category = 'High'
        quality_description = f'Excellent match quality (Q: {penalized_score:.1f})'
    elif penalized_score >= 4.0:
        quality_category = 'Medium'
        quality_description = f'Good match quality (Q: {penalized_score:.1f})'
    elif penalized_score >= 2.5:
        quality_category = 'Low'
        quality_description = f'Poor match quality (Q: {penalized_score:.1f})'
    else:
        quality_category = 'Very Low'
        quality_description = f'Very poor match quality (Q: {penalized_score:.1f})'
    
    # Add penalty information if applicable
    if penalty_factor < 1.0:
        quality_description += f' [Penalty applied: {penalty_factor:.2f} for {cluster_size} matches < 5]'
    
    return {
        'quality_category': quality_category,
        'quality_description': quality_description,
        'mean_top_5': top_5_mean,
        'penalized_score': penalized_score,
        'penalty_factor': penalty_factor,
        'cluster_size': cluster_size,
        'quality_metric': metric_name
    }


def _log_cluster_selection_details(assessment: Dict[str, Any]) -> None:
    """Log detailed information about cluster selection."""
    winning_info = assessment['winning_cluster_info']
    confidence = assessment['confidence_assessment']
    quality = assessment['quality_assessment']
    
    _LOGGER.info(f"🏆 NEW CLUSTER SELECTION METHOD RESULTS:")
    _LOGGER.info(f"   Winner: {winning_info['cluster_type']} cluster {winning_info['cluster_id']}")
    _LOGGER.info(f"   Cluster size: {winning_info['cluster_size']} templates")
    _LOGGER.info(f"   Top-5 mean: {winning_info['top_5_mean']:.3f}")
    _LOGGER.info(f"   Penalty factor: {winning_info['penalty_factor']:.3f}")
    _LOGGER.info(f"   Final score: {winning_info['penalized_score']:.3f}")
    try:
        share_pct = float(winning_info.get('top_1_share', 0.0)) * 100.0
        _LOGGER.info(f"   Top-1 share of weighted top-5: {share_pct:.1f}%")
    except Exception:
        pass
    
    _LOGGER.info(f"🔍 CONFIDENCE ASSESSMENT:")
    _LOGGER.info(f"   Confidence level: {confidence['confidence_level'].upper()}")
    _LOGGER.info(f"   {confidence['confidence_description']}")
    
    _LOGGER.info(f"📊 QUALITY ASSESSMENT:")
    _LOGGER.info(f"   Quality category: {quality['quality_category']}")
    _LOGGER.info(f"   {quality['quality_description']}")
    
    # Show top 3 clusters
    _LOGGER.info(f"🏅 TOP 3 CLUSTERS:")
    for i, cluster_info in enumerate(assessment['all_cluster_scores'][:3], 1):
        disqualified = " [DISQUALIFIED: below quality threshold]" if cluster_info['penalized_score'] == 0.0 and cluster_info['top_5_mean'] > 0 else ""
        _LOGGER.info(f"   {i}. {cluster_info['cluster_type']} (score: {cluster_info['penalized_score']:.3f}, "
                    f"size: {cluster_info['cluster_size']}, "
                    f"top-5 mean: {cluster_info['top_5_mean']:.3f}){disqualified}")





from PySide6 import QtWidgets

from snid_sage.interfaces.gui.components.widgets.flexible_number_input import (
    create_flexible_double_input,
)


def create_options(dialog, layout: QtWidgets.QVBoxLayout) -> None:
    desc = QtWidgets.QLabel("Apply edge tapering to prevent artifacts at spectrum boundaries.")
    desc.setWordWrap(True)
    desc.setStyleSheet("color: #64748b; font-size: 11pt; margin-bottom: 10px;")
    layout.addWidget(desc)

    apod_group = QtWidgets.QGroupBox("Apodization Configuration")
    apod_layout = QtWidgets.QVBoxLayout(apod_group)

    dialog.apodize_cb = QtWidgets.QCheckBox("Apply edge apodization")
    dialog.apodize_cb.setChecked(dialog.processing_params['apply_apodization'])
    dialog.apodize_cb.toggled.connect(lambda *_: _on_apodize_toggled(dialog))
    apod_layout.addWidget(dialog.apodize_cb)

    param_layout = QtWidgets.QHBoxLayout()
    param_layout.addWidget(QtWidgets.QLabel("Edge fraction:"))
    dialog.apod_percent_spin = create_flexible_double_input(min_val=1.0, max_val=50.0, suffix=" %", default=10.0)
    dialog.apod_percent_spin.setValue(dialog.processing_params['apod_percent'])
    dialog.apod_percent_spin.valueChanged.connect(lambda *_: _on_apod_percent_changed(dialog))
    param_layout.addWidget(dialog.apod_percent_spin)
    param_layout.addStretch()
    apod_layout.addLayout(param_layout)
    layout.addWidget(apod_group)

    info_text = QtWidgets.QLabel(
        "Apodization applies a smooth taper to the spectrum edges, preventing discontinuities that could affect Fourier-based analysis."
    )
    info_text.setWordWrap(True)
    info_text.setStyleSheet("color: #64748b; font-size: 10pt; font-style: italic;")
    layout.addWidget(info_text)


def apply_step(dialog) -> None:
    apply_apodization = False
    percent = 10.0
    try:
        if hasattr(dialog, 'apodize_cb') and dialog.apodize_cb is not None:
            apply_apodization = bool(dialog.apodize_cb.isChecked())
        if hasattr(dialog, 'apod_percent_spin') and dialog.apod_percent_spin is not None:
            percent = float(dialog.apod_percent_spin.value())
    except Exception:
        pass
    if apply_apodization:
        dialog.preview_calculator.apply_step("apodization", percent=percent, step_index=4)


def calculate_preview(dialog):
    try:
        if hasattr(dialog, 'apodize_cb') and dialog.apodize_cb.isChecked():
            percent = dialog.apod_percent_spin.value() if hasattr(dialog, 'apod_percent_spin') else 10.0
            if 0 <= percent <= 50:
                return dialog.preview_calculator.preview_step("apodization", percent=percent)
    except Exception:
        pass
    return dialog.preview_calculator.get_current_state()


def _on_apodize_toggled(dialog):
    dialog.processing_params['apply_apodization'] = bool(dialog.apodize_cb.isChecked()) if hasattr(dialog, 'apodize_cb') else False
    dialog._update_preview()


def _on_apod_percent_changed(dialog):
    try:
        if hasattr(dialog, 'apod_percent_spin') and dialog.apod_percent_spin is not None:
            dialog.processing_params['apod_percent'] = float(dialog.apod_percent_spin.value())
    except Exception:
        pass
    dialog._update_preview()





"""
Preprocessing Features Package
=============================

This package contains preprocessing-specific functionality for the SNID GUI,
including core preprocessing classes and utilities.

Components:
  (PySide6) PySide6PreviewCalculator for preview operations
"""

__all__ = []

# Preprocessing Features Module 



"""
SNID SAGE - Button Palette Demo (PySide6)
=========================================

A developer/demo dialog to preview and tweak the colors used for dialog buttons
and compare them with the main GUI workflow button palette. It supports:
- Live preview of dialog button types (apply, cancel, utility, etc.)
- Quick editing of base colors (auto-generates hover/pressed variants)
- Reset to defaults or to main GUI workflow palette
- Export current palette to JSON (clipboard)

This does NOT change global theme permanently; it's a preview/testing tool.
"""

from typing import Dict, List
import json

import PySide6.QtCore as QtCore
import PySide6.QtGui as QtGui
import PySide6.QtWidgets as QtWidgets

try:
    from snid_sage.shared.utils.logging import get_logger
    _LOGGER = get_logger('gui.button_palette_demo')
except Exception:  # pragma: no cover
    import logging
    _LOGGER = logging.getLogger('gui.button_palette_demo')


# Enhanced dialog button manager (provides the button type system and colors)
from snid_sage.interfaces.gui.utils.enhanced_dialog_button_manager import (
    EnhancedDialogButtonManager,
)

# Theme manager (for main GUI workflow/button palette reference)
try:
    from snid_sage.interfaces.gui.utils.pyside6_theme_manager import (
        get_pyside6_theme_manager,
    )
    THEME_MANAGER_AVAILABLE = True
except Exception:
    THEME_MANAGER_AVAILABLE = False


class PySide6ButtonPaletteDemoDialog(QtWidgets.QDialog):
    """Dialog to preview and tweak dialog button colors with live updates."""

    # Button types we expose for editing (keys must match EnhancedDialogButtonManager.BUTTON_COLORS base keys)
    EDITABLE_TYPES: List[str] = [
        'apply', 'secondary', 'cancel', 'utility', 'info', 'reset', 'navigation', 'neutral', 'accent'
    ]

    WORKFLOW_KEYS: List[str] = [
        'btn_load', 'btn_preprocessing', 'btn_redshift', 'btn_analysis', 'btn_advanced',
        'btn_ai', 'btn_settings', 'btn_reset'
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Button Palette Demo")
        self.resize(980, 640)
        self.setModal(False)

        # Make a local manager to help with lighten/darken and re-style
        self.button_manager = EnhancedDialogButtonManager()

        # Keep a deep copy of original colors so we can reset
        self.original_colors: Dict[str, str] = dict(EnhancedDialogButtonManager.BUTTON_COLORS)

        # Build UI
        self._build_ui()
        self._populate_workflow_palette()
        self._register_demo_buttons()

    # ---------- UI construction ----------
    def _build_ui(self) -> None:
        main_layout = QtWidgets.QHBoxLayout(self)
        main_layout.setContentsMargins(12, 12, 12, 12)
        main_layout.setSpacing(12)

        # Left: live previews
        left_splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        main_layout.addWidget(left_splitter, 2)

        # Group 1: Dialog button types preview
        types_group = QtWidgets.QGroupBox("Dialog Button Types (live preview)")
        types_layout = QtWidgets.QGridLayout(types_group)
        types_layout.setContentsMargins(10, 10, 10, 10)
        types_layout.setHorizontalSpacing(8)
        types_layout.setVerticalSpacing(8)

        self.preview_buttons: Dict[str, QtWidgets.QPushButton] = {}
        row = 0
        for i, btn_type in enumerate(self.EDITABLE_TYPES):
            label = QtWidgets.QLabel(btn_type.title())
            types_layout.addWidget(label, row, 0)

            # Normal-size button
            normal_btn = QtWidgets.QPushButton("Sample")
            normal_btn.setObjectName(f"demo_{btn_type}_normal")
            self.preview_buttons[f"{btn_type}_normal"] = normal_btn
            types_layout.addWidget(normal_btn, row, 1)

            # Small-size button
            small_btn = QtWidgets.QPushButton("Small")
            small_btn.setObjectName(f"demo_{btn_type}_small")
            self.preview_buttons[f"{btn_type}_small"] = small_btn
            types_layout.addWidget(small_btn, row, 2)

            row += 1

        left_splitter.addWidget(types_group)

        # Group 2: Main GUI workflow colors (reference)
        workflow_group = QtWidgets.QGroupBox("Main GUI Workflow Colors (reference)")
        workflow_layout = QtWidgets.QGridLayout(workflow_group)
        workflow_layout.setContentsMargins(10, 10, 10, 10)
        workflow_layout.setHorizontalSpacing(8)
        workflow_layout.setVerticalSpacing(8)

        self.workflow_color_labels: Dict[str, QtWidgets.QLabel] = {}
        self.workflow_color_buttons: Dict[str, QtWidgets.QPushButton] = {}
        for idx, key in enumerate(self.WORKFLOW_KEYS):
            text = key.replace('btn_', '').title()
            lbl = QtWidgets.QLabel(text)
            workflow_layout.addWidget(lbl, idx, 0)

            # Color swatch button (disabled, just displays color)
            swatch = QtWidgets.QPushButton()
            swatch.setEnabled(False)
            swatch.setFixedWidth(140)
            self.workflow_color_buttons[key] = swatch
            workflow_layout.addWidget(swatch, idx, 1)

            # Hex label
            hex_lbl = QtWidgets.QLabel("")
            self.workflow_color_labels[key] = hex_lbl
            workflow_layout.addWidget(hex_lbl, idx, 2)

        left_splitter.addWidget(workflow_group)

        # Right: controls to edit colors
        right_panel = QtWidgets.QWidget()
        right_layout = QtWidgets.QVBoxLayout(right_panel)
        right_layout.setContentsMargins(10, 10, 10, 10)
        right_layout.setSpacing(8)
        main_layout.addWidget(right_panel, 1)

        # Editor table
        self.editor_table = QtWidgets.QTableWidget()
        self.editor_table.setColumnCount(4)
        self.editor_table.setHorizontalHeaderLabels(["Type", "Base Color", "Pick", "Apply"])
        self.editor_table.horizontalHeader().setStretchLastSection(True)
        self.editor_table.setRowCount(len(self.EDITABLE_TYPES))

        self.hex_inputs: Dict[str, QtWidgets.QLineEdit] = {}
        for row_idx, t in enumerate(self.EDITABLE_TYPES):
            self.editor_table.setItem(row_idx, 0, QtWidgets.QTableWidgetItem(t))

            # Hex input
            hex_edit = QtWidgets.QLineEdit(self._get_base_color_hex(t))
            hex_edit.setMaxLength(7)
            hex_edit.setPlaceholderText("#RRGGBB")
            self.hex_inputs[t] = hex_edit
            self.editor_table.setCellWidget(row_idx, 1, hex_edit)

            # Pick button
            pick_btn = QtWidgets.QPushButton("Pick…")
            pick_btn.clicked.connect(lambda _, tp=t: self._pick_color(tp))
            self.editor_table.setCellWidget(row_idx, 2, pick_btn)

            # Apply button
            apply_btn = QtWidgets.QPushButton("Apply")
            apply_btn.clicked.connect(lambda _, tp=t: self._apply_color_for_type(tp))
            self.editor_table.setCellWidget(row_idx, 3, apply_btn)

        right_layout.addWidget(self.editor_table)

        # Interactive color picker
        self._build_color_picker(right_layout)

        # Actions
        actions_layout = QtWidgets.QHBoxLayout()
        right_layout.addLayout(actions_layout)

        self.reset_defaults_btn = QtWidgets.QPushButton("Reset to Defaults")
        self.reset_defaults_btn.clicked.connect(self._reset_to_defaults)
        actions_layout.addWidget(self.reset_defaults_btn)

        self.reset_main_gui_btn = QtWidgets.QPushButton("Reset to Main GUI Palette")
        self.reset_main_gui_btn.clicked.connect(self._reset_to_main_gui)
        actions_layout.addWidget(self.reset_main_gui_btn)

        actions_layout.addStretch()

        export_btn = QtWidgets.QPushButton("Copy JSON to Clipboard")
        export_btn.clicked.connect(self._export_palette_json)
        right_layout.addWidget(export_btn)

        # Footer info
        note = QtWidgets.QLabel("This tool previews colors only. It does not persist changes.")
        note.setStyleSheet("color: #64748b; font-style: italic;")
        right_layout.addWidget(note)

    # ---------- Helpers ----------
    def _register_demo_buttons(self) -> None:
        """Register all preview buttons with the manager using current colors."""
        for key, btn in self.preview_buttons.items():
            btn_type = key.split('_')[0]
            size = 'small' if key.endswith('small') else 'normal'
            self.button_manager.register_button(btn, button_type=btn_type, custom_config={'size_class': size})

    # ---------- Interactive picker ----------
    def _build_color_picker(self, parent_layout: QtWidgets.QVBoxLayout) -> None:
        group = QtWidgets.QGroupBox("Interactive Color Picker")
        layout = QtWidgets.QVBoxLayout(group)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(8)

        # Target type selector
        selector_row = QtWidgets.QHBoxLayout()
        selector_row.addWidget(QtWidgets.QLabel("Target Type:"))
        self.picker_type_combo = QtWidgets.QComboBox()
        self.picker_type_combo.addItems(self.EDITABLE_TYPES)
        self.picker_type_combo.currentTextChanged.connect(self._sync_picker_to_type)
        selector_row.addWidget(self.picker_type_combo)
        selector_row.addStretch()
        layout.addLayout(selector_row)

        # Hue slider
        hue_row = QtWidgets.QHBoxLayout()
        self.hue_label = QtWidgets.QLabel("Hue: 0")
        self.hue_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.hue_slider.setRange(0, 359)
        self.hue_slider.valueChanged.connect(self._on_hue_changed)
        hue_row.addWidget(self.hue_label)
        hue_row.addWidget(self.hue_slider)
        layout.addLayout(hue_row)

        # SV selector + preview
        sv_row = QtWidgets.QHBoxLayout()
        self.sv_selector = _ColorSVSelector()
        self.sv_selector.colorChanged.connect(self._on_picker_color_changed)
        sv_row.addWidget(self.sv_selector, 1)

        preview_col = QtWidgets.QVBoxLayout()
        preview_col.addWidget(QtWidgets.QLabel("Selected:"))
        self.picker_preview = QtWidgets.QLabel()
        self.picker_preview.setFixedSize(56, 56)
        self.picker_preview.setStyleSheet("border: 1px solid #cbd5e1; border-radius: 4px;")
        preview_col.addWidget(self.picker_preview)
        preview_col.addStretch()
        sv_row.addLayout(preview_col)
        layout.addLayout(sv_row)

        # Apply button
        apply_row = QtWidgets.QHBoxLayout()
        self.picker_apply_btn = QtWidgets.QPushButton("Apply to Selected Type")
        self.picker_apply_btn.clicked.connect(self._apply_picker_to_type)
        apply_row.addWidget(self.picker_apply_btn)
        apply_row.addStretch()
        layout.addLayout(apply_row)

        parent_layout.addWidget(group)

        # Initialize picker from current type
        self._sync_picker_to_type(self.picker_type_combo.currentText())

    def _populate_workflow_palette(self) -> None:
        """Populate the main workflow color swatches from the theme manager if available."""
        if not THEME_MANAGER_AVAILABLE:
            return
        tm = get_pyside6_theme_manager()
        colors = tm.get_all_colors()
        for key in self.WORKFLOW_KEYS:
            hex_color = colors.get(key, '#999999')
            self._set_swatch(self.workflow_color_buttons[key], hex_color)
            self.workflow_color_labels[key].setText(hex_color)

    def _set_swatch(self, button: QtWidgets.QPushButton, hex_color: str) -> None:
        button.setStyleSheet(
            f"background-color: {hex_color}; border: 1px solid #cbd5e1; border-radius: 4px; min-height: 22px;"
        )

    def _get_base_color_hex(self, btn_type: str) -> str:
        """Get base color hex for a type from the manager's palette."""
        return EnhancedDialogButtonManager.BUTTON_COLORS.get(btn_type, '#888888')

    def _lighten(self, hex_color: str, factor: float) -> str:
        # Reuse manager's logic via a temporary method call
        return self.button_manager._lighten_color(hex_color, factor)

    def _darken(self, hex_color: str, factor: float) -> str:
        return self.button_manager._darken_color(hex_color, factor)

    # ---------- Actions ----------
    def _pick_color(self, btn_type: str) -> None:
        current_hex = self.hex_inputs[btn_type].text().strip() or self._get_base_color_hex(btn_type)
        color = QtGui.QColor(current_hex)
        if not color.isValid():
            color = QtGui.QColor('#888888')
        chosen = QtWidgets.QColorDialog.getColor(color, self, f"Pick color for {btn_type}")
        if chosen.isValid():
            new_hex = chosen.name().lower()
            self.hex_inputs[btn_type].setText(new_hex)
            self._apply_color_for_type(btn_type)

    def _on_hue_changed(self, value: int) -> None:
        self.hue_label.setText(f"Hue: {value}")
        self.sv_selector.setHue(value)

    def _on_picker_color_changed(self, color: QtGui.QColor) -> None:
        # Update preview swatch
        self.picker_preview.setStyleSheet(
            f"background-color: {color.name()}; border: 1px solid #cbd5e1; border-radius: 4px;"
        )
        # Live-update hex input for selected type
        btn_type = self.picker_type_combo.currentText()
        self.hex_inputs[btn_type].setText(color.name())
        # Live-apply to preview buttons
        self._apply_color_for_type(btn_type)

    def _apply_picker_to_type(self) -> None:
        btn_type = self.picker_type_combo.currentText()
        hex_color = self.sv_selector.currentColor().name()
        self.hex_inputs[btn_type].setText(hex_color)
        self._apply_color_for_type(btn_type)

    def _sync_picker_to_type(self, btn_type: str) -> None:
        hex_color = self._get_base_color_hex(btn_type)
        qc = QtGui.QColor(hex_color)
        # Handle gray (no hue) case gracefully
        hue = qc.hue() if qc.hue() >= 0 else 0
        sat = qc.saturationF() if qc.isValid() else 0.0
        val = qc.valueF() if qc.isValid() else 1.0
        self.hue_slider.blockSignals(True)
        self.hue_slider.setValue(hue)
        self.hue_slider.blockSignals(False)
        self.sv_selector.blockSignals(True)
        self.sv_selector.setHue(hue)
        self.sv_selector.setSV(sat, val)
        self.sv_selector.blockSignals(False)
        self._on_picker_color_changed(self.sv_selector.currentColor())

    def _apply_color_for_type(self, btn_type: str) -> None:
        base_hex = self.hex_inputs[btn_type].text().strip()
        if not base_hex.startswith('#') or len(base_hex) != 7:
            QtWidgets.QMessageBox.warning(self, "Invalid Color", f"Enter a valid hex color for {btn_type} (e.g., #3b82f6)")
            return

        # Update global palette for the type and its hover/pressed variants
        EnhancedDialogButtonManager.BUTTON_COLORS[btn_type] = base_hex
        EnhancedDialogButtonManager.BUTTON_COLORS[f"{btn_type}_hover"] = self._darken(base_hex, 0.10)
        EnhancedDialogButtonManager.BUTTON_COLORS[f"{btn_type}_pressed"] = self._darken(base_hex, 0.18)

        # Re-apply styling to preview buttons of this type
        for key, btn in self.preview_buttons.items():
            if key.startswith(btn_type + "_"):
                # Trigger a restyle by toggling enabled state
                self.button_manager.update_button_state(btn, enabled=True)

    def _reset_to_defaults(self) -> None:
        EnhancedDialogButtonManager.BUTTON_COLORS.update(self.original_colors)
        # Refresh editor inputs
        for t in self.EDITABLE_TYPES:
            self.hex_inputs[t].setText(self._get_base_color_hex(t))
            self._apply_color_for_type(t)

    def _reset_to_main_gui(self) -> None:
        """Reset dialog button types to colors inspired by main GUI workflow palette."""
        # Map dialog types to workflow palette keys
        if THEME_MANAGER_AVAILABLE:
            tm = get_pyside6_theme_manager()
            c = tm.get_all_colors()
            mapping = {
                'apply': c.get('btn_success', '#10b981'),
                'secondary': c.get('btn_analysis', '#BC5090'),
                'cancel': c.get('btn_reset', '#A65965'),
                'utility': c.get('btn_advanced', '#58508D'),
                'info': c.get('btn_primary', '#3b82f6'),
                'reset': c.get('btn_reset', '#A65965'),
                'navigation': c.get('btn_preprocessing', '#FF6361'),
                'neutral': c.get('btn_ai', '#003F5C'),
                'accent': c.get('btn_success', '#10b981'),
            }
        else:
            # Fallback constants matching current defaults
            mapping = {
                'apply': '#22c55e',
                'secondary': '#BC5090',
                'cancel': '#A65965',
                'utility': '#58508D',
                'info': '#6366f1',
                'reset': '#6E6E6E',
                'navigation': '#FF6361',
                'neutral': '#003F5C',
                'accent': '#22c55e',
            }

        for t, hex_color in mapping.items():
            self.hex_inputs[t].setText(hex_color)
            self._apply_color_for_type(t)

    def _export_palette_json(self) -> None:
        """Copy current base colors for editable types as JSON to clipboard."""
        export = {t: self._get_base_color_hex(t) for t in self.EDITABLE_TYPES}
        json_text = json.dumps(export, indent=2)
        QtWidgets.QApplication.clipboard().setText(json_text)
        QtWidgets.QMessageBox.information(self, "Exported", "Current palette copied to clipboard as JSON")


def show_button_palette_demo(parent=None) -> PySide6ButtonPaletteDemoDialog:
    dialog = PySide6ButtonPaletteDemoDialog(parent)
    dialog.show()
    return dialog


class _ColorSVSelector(QtWidgets.QWidget):
    """Simple Saturation/Value selector with fixed hue and live color reporting."""
    colorChanged = QtCore.Signal(QtGui.QColor)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumSize(220, 220)
        self._hue = 0  # 0..359
        self._s = 1.0  # 0..1
        self._v = 1.0  # 0..1
        self._cache_img: QtGui.QImage | None = None
        self._cache_hue: int | None = None
        # Do not track passive mouse moves; only update on click/drag
        self.setMouseTracking(False)

    def setHue(self, hue: int) -> None:
        hue = max(0, min(359, int(hue)))
        if hue != self._hue:
            self._hue = hue
            self._cache_img = None
            self.update()
            self.colorChanged.emit(self.currentColor())

    def setSV(self, s: float, v: float) -> None:
        s = max(0.0, min(1.0, float(s)))
        v = max(0.0, min(1.0, float(v)))
        if s != self._s or v != self._v:
            self._s, self._v = s, v
            self.update()
            self.colorChanged.emit(self.currentColor())

    def currentColor(self) -> QtGui.QColor:
        return QtGui.QColor.fromHsv(self._hue, int(self._s * 255), int(self._v * 255))

    def _ensure_cache(self) -> None:
        if self._cache_img is not None and self._cache_hue == self._hue and \
           self._cache_img.size() == self.size():
            return
        w, h = self.width(), self.height()
        img = QtGui.QImage(w, h, QtGui.QImage.Format_RGB32)
        for y in range(h):
            v = 1.0 - (y / max(1, h - 1))
            for x in range(w):
                s = x / max(1, w - 1)
                qc = QtGui.QColor.fromHsv(self._hue, int(s * 255), int(v * 255))
                img.setPixel(x, y, qc.rgb())
        self._cache_img = img
        self._cache_hue = self._hue

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        self._ensure_cache()
        painter = QtGui.QPainter(self)
        if self._cache_img is not None:
            painter.drawImage(0, 0, self._cache_img)
        # Draw selection marker
        x = self._s * (self.width() - 1)
        y = (1.0 - self._v) * (self.height() - 1)
        pen = QtGui.QPen(QtGui.QColor('#ffffff'))
        pen.setWidth(2)
        painter.setPen(pen)
        painter.drawEllipse(QtCore.QPointF(x, y), 6, 6)
        pen = QtGui.QPen(QtGui.QColor('#000000'))
        pen.setWidth(1)
        painter.setPen(pen)
        painter.drawEllipse(QtCore.QPointF(x, y), 7, 7)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        self._update_from_pos(event.position())

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        # Only update while dragging with left button
        if event.buttons() & QtCore.Qt.LeftButton:
            self._update_from_pos(event.position())

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        self._cache_img = None
        super().resizeEvent(event)

    def _update_from_pos(self, pos: QtCore.QPointF) -> None:
        w, h = max(1, self.width() - 1), max(1, self.height() - 1)
        s = max(0.0, min(1.0, pos.x() / w))
        v = max(0.0, min(1.0, 1.0 - (pos.y() / h)))
        if s != self._s or v != self._v:
            self._s, self._v = s, v
            self.update()
            self.colorChanged.emit(self.currentColor())





"""
SNID SAGE - Enhanced Button Manager
====================================

Provides smooth visual feedback for main workflow buttons including:
- Hover animations with color transitions
- Click animations with subtle scale effects
- Professional visual feedback using existing theme colors
- No conflicts with existing workflow and theme systems

Developed by Assistant for SNID SAGE
"""

from typing import Dict, Optional
import PySide6.QtCore as QtCore
import PySide6.QtGui as QtGui
import PySide6.QtWidgets as QtWidgets

# Import logging
try:
    from snid_sage.shared.utils.logging import get_logger
    _LOGGER = get_logger('gui.enhanced_buttons')
except ImportError:
    import logging
    _LOGGER = logging.getLogger('gui.enhanced_buttons')


class EnhancedButtonManager(QtCore.QObject):
    """
    Manager for enhanced button visual feedback
    
    This class adds smooth hover and click animations to workflow buttons
    while preserving all existing functionality and styling systems.
    Uses predefined colors for optimal performance.
    """
    
    # Predefined color states for all button types (faster than runtime calculations)
    BUTTON_COLOR_STATES = {
        'load': {
            'base': '#6E6E6E',
            'hover': '#656565', 
            'pressed': '#585858'
        },
        'redshift': {
            'base': '#FFA600',
            'hover': '#E89500',
            'pressed': '#CC8400'
        },
        'preprocessing': {
            'base': '#FF6361',
            'hover': '#E85A58',
            'pressed': '#CC514F'
        },
        'analysis': {
            'base': '#BC5090',
            'hover': '#A84981',
            'pressed': '#944272'
        },
        'advanced': {
            'base': '#58508D',
            'hover': '#4F477E',
            'pressed': '#463E6F'
        },
        'ai': {
            'base': '#003F5C',
            'hover': '#003A54',
            'pressed': '#00344B'
        },
        'settings': {
            'base': '#7A8585',
            'hover': '#6E7777',
            'pressed': '#626969'
        },
        'reset': {
            'base': '#A65965',
            'hover': '#95515C',
            'pressed': '#844953'
        },
        'info': {
            'base': '#3b82f6',
            'hover': '#2563eb',
            'pressed': '#1d4ed8'
        },
        'neutral': {
            'base': '#9ca3af',
            'hover': '#6b7280',
            'pressed': '#4b5563'
        }
    }
    
    def __init__(self, theme_manager):
        """Initialize the enhanced button manager"""
        super().__init__()
        self.theme_manager = theme_manager
        self.enhanced_buttons: Dict[str, QtWidgets.QPushButton] = {}
        self.animations: Dict[str, QtCore.QPropertyAnimation] = {}
        
        _LOGGER.info("Enhanced Button Manager initialized")
    
    def register_button(self, button: QtWidgets.QPushButton, color_type: str):
        """Register a button for enhanced feedback"""
        if not button:
            _LOGGER.warning("Cannot register None button")
            return
            
        object_name = button.objectName()
        if not object_name:
            # Generate a fallback object name if none exists
            object_name = f"enhanced_btn_{id(button)}"
            button.setObjectName(object_name)
            _LOGGER.debug(f"Generated object name for button: {object_name}")
        
        # Store button and its color type
        self.enhanced_buttons[object_name] = button
        button.setProperty("enhanced_color_type", color_type)
        button.setProperty("is_enhanced", True)
        
        # Install event filter for enhanced interactions
        button.installEventFilter(self)
        
        # Set cursor to pointing hand
        button.setCursor(QtCore.Qt.PointingHandCursor)
        
        _LOGGER.debug(f"Button {object_name} registered for enhanced feedback with color type: {color_type}")
    
    def eventFilter(self, obj: QtCore.QObject, event: QtCore.QEvent) -> bool:
        """Handle enhanced button interactions"""
        if not isinstance(obj, QtWidgets.QPushButton):
            return super().eventFilter(obj, event)
        
        if not obj.property("is_enhanced"):
            return super().eventFilter(obj, event)
        
        # Only process events for enabled buttons
        if not obj.isEnabled():
            return super().eventFilter(obj, event)
        
        try:
            if event.type() == QtCore.QEvent.Enter:
                self._animate_hover_enter(obj)
            elif event.type() == QtCore.QEvent.Leave:
                self._animate_hover_leave(obj)
            elif event.type() == QtCore.QEvent.MouseButtonPress:
                if event.button() == QtCore.Qt.LeftButton:
                    _LOGGER.debug(f"Button press animation triggered: {obj.objectName()} (left click)")
                    self._animate_button_press(obj)
                elif event.button() == QtCore.Qt.RightButton:
                    # Only show enhanced feedback for right-clicks on buttons that have right-click functionality
                    if self._button_has_right_click_functionality(obj):
                        _LOGGER.debug(f"Button press animation triggered: {obj.objectName()} (right click)")
                        self._animate_button_press(obj)
            elif event.type() == QtCore.QEvent.MouseButtonRelease:
                if event.button() == QtCore.Qt.LeftButton:
                    _LOGGER.debug(f"Button release animation triggered: {obj.objectName()} (left click)")
                    self._animate_button_release(obj)
                elif event.button() == QtCore.Qt.RightButton:
                    # Only show enhanced feedback for right-clicks on buttons that have right-click functionality
                    if self._button_has_right_click_functionality(obj):
                        _LOGGER.debug(f"Button release animation triggered: {obj.objectName()} (right click)")
                        self._animate_button_release(obj)
        except Exception as e:
            _LOGGER.error(f"Error in enhanced button event handling: {e}")
        
        return super().eventFilter(obj, event)
    
    def _button_has_right_click_functionality(self, button: QtWidgets.QPushButton) -> bool:
        """Check if a button has right-click functionality that should show enhanced feedback"""
        object_name = button.objectName()
        
        # Only Preprocessing and Analysis buttons have right-click functionality
        right_click_buttons = {
            'unified_preprocessing_btn',      # Preprocessing button
            'unified_analysis_btn',           # Analysis button
            'preprocessing_btn',              # Alternative name
            'analysis_btn',                   # Alternative name
        }
        
        return object_name in right_click_buttons
    
    def _is_small_button(self, object_name: str) -> bool:
        """Determine if a button should use small dimensions (24px height)"""
        small_buttons = {
            'unified_prev_btn',
            'unified_next_btn',
            'unified_cluster_summary_btn',
            'unified_gmm_btn',
            'unified_redshift_age_btn',
            'unified_subtype_proportions_btn',
            'unified_reset_btn',
            'unified_settings_btn',
            'unified_info_btn',
        }
        return object_name in small_buttons
    
    def _animate_hover_enter(self, button: QtWidgets.QPushButton):
        """Animate button on hover enter"""
        color_type = button.property("enhanced_color_type")
        if not color_type:
            return
        
        # Check if this is the cluster summary button and it's currently blinking
        # If so, don't apply hover effects to avoid size conflicts
        object_name = button.objectName()
        if object_name == "unified_cluster_summary_btn":
            if hasattr(self, '_workflow_manager_ref') and self._workflow_manager_ref:
                main_window = self._workflow_manager_ref.main_window
                if hasattr(main_window, 'cluster_summary_blinking') and main_window.cluster_summary_blinking:
                    return  # Skip hover effects while blinking
        
        base_color = self._get_color_for_type(color_type)
        if not base_color:
            return
        
        # Get predefined hover color for optimal performance (no runtime calculations)
        color_states = self.BUTTON_COLOR_STATES.get(color_type, self.BUTTON_COLOR_STATES['neutral'])
        hover_color_1 = color_states['hover']  # Predefined hover color
        hover_color_2 = base_color  # Original bottom for gradient
        
        object_name = button.objectName()
        
        # Determine button size based on object name
        if self._is_small_button(object_name):
            # Small buttons (navigation, analysis, utility) - use smaller dimensions
            if "unified_prev_btn" in object_name or "unified_next_btn" in object_name:
                font_size = "14pt"  # Navigation arrows
            elif "unified_reset_btn" in object_name or "unified_settings_btn" in object_name:
                font_size = "9pt"   # Utility buttons
            elif "unified_info_btn" in object_name:
                font_size = "14px"  # Info button (uses px like in its inline style)
            elif "unified_gmm_btn" in object_name or "unified_redshift_age_btn" in object_name or "unified_subtype_proportions_btn" in object_name:
                font_size = "14pt"  # Larger emoji buttons
            else:
                font_size = "12pt"  # Default small buttons (cluster summary)
            padding = "2px 4px"
            min_height = "24px"
        else:
            # Main workflow buttons - use standard dimensions
            font_size = "11pt"
            padding = "6px 12px"
            min_height = "32px"
        
        enhanced_style = f"""
        QPushButton#{object_name} {{
            background: qlineargradient(x1:0, y1:0, x2:0, y2:1,
                stop:0 {hover_color_1},
                stop:1 {hover_color_2});
            color: white;
            border: 2px solid {base_color};
            border-radius: 4px;
            font-weight: bold;
            font-size: {font_size};
            padding: {padding};
            min-height: {min_height};
        }}
        """
        
        button.setStyleSheet(enhanced_style)
    
    def _animate_hover_leave(self, button: QtWidgets.QPushButton):
        """Animate button on hover leave - restore workflow manager styling"""
        # Let the workflow manager handle the styling restoration
        # This ensures consistency with the existing system
        object_name = button.objectName()
        color_type = button.property("enhanced_color_type")
        
        # Check if this is the cluster summary button and it's currently blinking
        # If so, don't apply hover leave effects to avoid size conflicts
        if object_name == "unified_cluster_summary_btn":
            if hasattr(self, '_workflow_manager_ref') and self._workflow_manager_ref:
                main_window = self._workflow_manager_ref.main_window
                if hasattr(main_window, 'cluster_summary_blinking') and main_window.cluster_summary_blinking:
                    return  # Skip hover leave effects while blinking
        
        if color_type and hasattr(self, '_workflow_manager_ref'):
            # Use workflow manager's styling with appropriate dimensions
            base_color = self._get_color_for_type(color_type)
            if not base_color:
                return
                
            # Determine button size based on object name
            if self._is_small_button(object_name):
                # Small buttons (navigation, analysis, utility) - use smaller dimensions
                if "unified_prev_btn" in object_name or "unified_next_btn" in object_name:
                    font_size = "14pt"  # Navigation arrows
                elif "unified_reset_btn" in object_name or "unified_settings_btn" in object_name:
                    font_size = "9pt"   # Utility buttons
                elif "unified_gmm_btn" in object_name or "unified_redshift_age_btn" in object_name or "unified_subtype_proportions_btn" in object_name:
                    font_size = "14pt"  # Larger emoji buttons
                else:
                    font_size = "12pt"  # Default small buttons (cluster summary)
                padding = "2px 4px"
                min_height = "24px"
            else:
                # Main workflow buttons - use standard dimensions
                font_size = "11pt"
                padding = "6px 12px"
                min_height = "32px"
            
            basic_style = f"""
            QPushButton#{object_name} {{
                background: {base_color} !important;
                color: white !important;
                border: 2px solid {base_color} !important;
                border-radius: 4px;
                font-weight: bold;
                font-size: {font_size};
                padding: {padding};
                min-height: {min_height};
            }}
            """
            button.setStyleSheet(basic_style)
        else:
            # Fallback to basic styling with correct dimensions
            base_color = self._get_color_for_type(color_type)
            if not base_color:
                return
            
            # Determine button size
            if self._is_small_button(object_name):
                if "unified_prev_btn" in object_name or "unified_next_btn" in object_name:
                    font_size = "14pt"  # Navigation arrows
                elif "unified_reset_btn" in object_name or "unified_settings_btn" in object_name:
                    font_size = "9pt"   # Utility buttons
                elif "unified_gmm_btn" in object_name or "unified_redshift_age_btn" in object_name or "unified_subtype_proportions_btn" in object_name:
                    font_size = "14pt"  # Larger emoji buttons
                else:
                    font_size = "12pt"  # Default small buttons (cluster summary)
                padding = "2px 4px"
                min_height = "24px"
            else:
                font_size = "11pt"
                padding = "6px 12px"
                min_height = "32px"
            
            basic_style = f"""
            QPushButton#{object_name} {{
                background: {base_color};
                color: white;
                border: 2px solid {base_color};
                border-radius: 4px;
                font-weight: bold;
                font-size: {font_size};
                padding: {padding};
                min-height: {min_height};
            }}
            """
            button.setStyleSheet(basic_style)
    
    def _animate_button_press(self, button: QtWidgets.QPushButton):
        """Animate button press with subtle scale effect"""
        color_type = button.property("enhanced_color_type")
        if not color_type:
            return
        
        base_color = self._get_color_for_type(color_type)
        if not base_color:
            return
        
        # Create subtle scale animation
        object_name = button.objectName()
        animation_key = f"{object_name}_press"
        
        # Clean up any existing animation
        if animation_key in self.animations:
            self.animations[animation_key].stop()
            del self.animations[animation_key]
        
        # No geometry animation to keep button shape consistent
        # We'll rely on color changes only for visual feedback
        
        # Enhanced pressed styling - keep same shape but change colors only
        # Use predefined pressed colors for optimal performance and consistency
        color_states = self.BUTTON_COLOR_STATES.get(color_type, self.BUTTON_COLOR_STATES['neutral'])
        pressed_color_1 = color_states['pressed']  # Predefined pressed color
        pressed_color_2 = color_states['hover']    # Slightly lighter gradient bottom
        
        # Determine button size based on object name
        if self._is_small_button(object_name):
            # Small buttons (navigation, analysis, utility) - use smaller dimensions
            if "unified_prev_btn" in object_name or "unified_next_btn" in object_name:
                font_size = "14pt"  # Navigation arrows
            elif "unified_reset_btn" in object_name or "unified_settings_btn" in object_name:
                font_size = "9pt"   # Utility buttons
            elif "unified_info_btn" in object_name:
                font_size = "14px"  # Info button (uses px like in its inline style)
            elif "unified_gmm_btn" in object_name or "unified_redshift_age_btn" in object_name or "unified_subtype_proportions_btn" in object_name:
                font_size = "14pt"  # Larger emoji buttons
            else:
                font_size = "12pt"  # Default small buttons (cluster summary)
            padding = "2px 4px"
            min_height = "24px"
        else:
            # Main workflow buttons - use standard dimensions
            font_size = "11pt"
            padding = "6px 12px"
            min_height = "32px"
        
        pressed_style = f"""
        QPushButton#{object_name} {{
            background: qlineargradient(x1:0, y1:0, x2:0, y2:1,
                stop:0 {pressed_color_1},
                stop:1 {pressed_color_2});
            color: white;
            border: 2px solid {base_color};
            border-radius: 4px;
            font-weight: bold;
            font-size: {font_size};
            padding: {padding};
            min-height: {min_height};
        }}
        """
        
        button.setStyleSheet(pressed_style)
    
    def _animate_button_release(self, button: QtWidgets.QPushButton):
        """Animate button release"""
        object_name = button.objectName()
        animation_key = f"{object_name}_release"
        press_animation_key = f"{object_name}_press"
        
        # Clean up press animation
        if press_animation_key in self.animations:
            self.animations[press_animation_key].stop()
            del self.animations[press_animation_key]
        
        # No geometry restoration needed since we don't change geometry anymore
        
        # Return to hover state
        self._animate_hover_enter(button)
    
    def _get_color_for_type(self, color_type: str) -> Optional[str]:
        """Get base color for button type using predefined colors"""
        color_states = self.BUTTON_COLOR_STATES.get(color_type, self.BUTTON_COLOR_STATES['neutral'])
        return color_states['base']
    
    # Color calculation methods removed - now using predefined colors for better performance
    
    def set_workflow_manager_reference(self, workflow_manager):
        """Set reference to workflow manager for proper style restoration"""
        self._workflow_manager_ref = workflow_manager
    
    def cleanup(self):
        """Clean up animations and event filters"""
        for animation in self.animations.values():
            animation.stop()
        self.animations.clear()
        
        for button in self.enhanced_buttons.values():
            button.removeEventFilter(self)
        
        self.enhanced_buttons.clear()
        _LOGGER.info("Enhanced Button Manager cleaned up")